"""Define types used in ReproEnv."""

import typing as ty

from mypy_extensions import TypedDict
from typing_extensions import Literal

# The path to the JSON file within the container, which contains the information of
# how the container was generated. The contents of the JSON file conform to the
# renderer schema, and the dictionary can be used to instantiate a new renderer object,
# which will reproduce the Dockerfile or Singularity recipe.
# Note that at the time of writing (04 March 2021), there is no version checking on the
# renderer dictionary, so compatibility is not guaranteed.
REPROENV_SPEC_FILE_IN_CONTAINER = "/.reproenv.json"

allowed_installation_methods = {"binaries", "source"}
installation_methods_type = Literal["binaries", "source"]

allowed_pkg_managers = {"apt", "yum"}
pkg_managers_type = Literal["apt", "yum"]

# Cross-reference the dictionary types below with the JSON schemas.


class _InstallationDependenciesType(TypedDict, total=False):
    """Dictionary of system dependencies, with package managers as keys. Different
    distributions use different package managers. For example, CentOS and Fedora use
    `yum`, and Debian and Ubuntu use `apt` and `dpkg`.
    """

    apt: ty.List[str]
    debs: ty.List[str]
    yum: ty.List[str]


class _TemplateArgumentsType(TypedDict):
    """Arguments (i.e., variables) that are used in the template."""

    required: ty.List[str]
    optional: ty.Mapping[str, str]


class _BaseTemplateType(TypedDict, total=False):
    """Keys common to both types of templates: binaries and source."""

    arguments: _TemplateArgumentsType
    env: ty.Mapping[str, str]
    dependencies: _InstallationDependenciesType
    instructions: str


class _SourceTemplateType(_BaseTemplateType):
    """Template that defines how to install software by source."""

    pass


class _BinariesTemplateType(_BaseTemplateType):
    """Template that defines how to install software from pre-compiled binaries."""

    urls: ty.Mapping[str, str]


class TemplateType(TypedDict, total=False):
    """Dictionary that includes a template for installing software from binaries, a
    template for installing software from source, or both.
    """

    name: str
    binaries: _BinariesTemplateType
    source: _SourceTemplateType
    alert: str


class _SingularityHeaderType(TypedDict, total=False):
    """Dictionary that defines the header of a Singularity recipe."""

    bootstrap: str
    from_: str



"""ReproEnv is a generic generator of Dockerfiles and Singularity files. It forms the
core of Neurodocker.
"""

from neurodocker.reproenv.renderers import DockerRenderer  # noqa: F401
from neurodocker.reproenv.renderers import SingularityRenderer  # noqa: F401
from neurodocker.reproenv.state import register_template  # noqa: F401
from neurodocker.reproenv.state import registered_templates  # noqa: F401
from neurodocker.reproenv.state import get_template  # noqa: F401
from neurodocker.reproenv.template import Template  # noqa: F401



"""Stateful objects in reproenv runtime."""

import copy
import json
import os
from pathlib import Path
import typing as ty

import jsonschema
import yaml

# The [C]SafeLoader will only load a subset of YAML, but that is fine for the
# purposes of this package.
try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeLoader  # type: ignore  # pragma: no cover

from neurodocker.reproenv.exceptions import RendererError
from neurodocker.reproenv.exceptions import TemplateError
from neurodocker.reproenv.exceptions import TemplateNotFound
from neurodocker.reproenv.types import TemplateType

_schemas_path = Path(__file__).parent / "schemas"

with (_schemas_path / "template.json").open("r") as f:
    _TEMPLATE_SCHEMA: ty.Dict = json.load(f)

with (_schemas_path / "renderer.json").open("r") as f:
    _RENDERER_SCHEMA: ty.Dict = json.load(f)


def _validate_template(template: TemplateType):
    """Validate template against JSON schema. Raise exception if invalid."""
    # TODO: should reproenv have a custom exception for invalid templates? probably
    try:
        jsonschema.validate(template, schema=_TEMPLATE_SCHEMA)
    except jsonschema.exceptions.ValidationError as e:
        raise TemplateError(f"Invalid template: {e.message}.") from e

    # TODO: Check that all variables in the instructions are listed in arguments.
    # something like https://stackoverflow.com/a/8284419/5666087
    # but that solution does not get attributes like foo in `self.foo`.
    # For now, this is taken care of in Renderer classes, but it would be good to move
    # that behavior here, so we can catch errors early.
    pass


def _validate_renderer(d):
    """Validate renderer dictionary against JSON schema. Raise exception if invalid."""
    try:
        jsonschema.validate(d, schema=_RENDERER_SCHEMA)
    except jsonschema.exceptions.ValidationError as e:
        raise RendererError(f"Invalid renderer dictionary: {e.message}.") from e


class _TemplateRegistry:
    """Object to hold templates in memory."""

    _templates: ty.Dict[str, TemplateType] = {}

    @classmethod
    def _reset(cls):
        """Clear all templates."""
        cls._templates = {}

    @classmethod
    def register(
        cls,
        path_or_template: ty.Union[str, os.PathLike, TemplateType],
        name: str = None,
    ) -> TemplateType:
        """Register a template. This will overwrite an existing template with the
        same name in the registry.

        The template is validated against reproenv's template JSON schema upon
        registration. An invalid template will raise an exception.

        Parameters
        ----------
        path_or_template : str, Path-like, or TemplateType
            Path to YAML file that defines the template, or the dictionary that
            represents the template.
        name : str
            Name of the template. This becomes the key to the template in the registry.
            The name is made lower-case. If `path_or_template` is a path, then `name`
            can be omitted and instead comes from `template["name"]`. If
            `path_or_template` is a `dict`, then `name` is required.
        """
        if isinstance(path_or_template, dict):
            if name is None:
                raise ValueError("`name` required when template is not a file")
            name = str(name)
            template = copy.deepcopy(path_or_template)
        else:
            path_or_template = Path(path_or_template)
            if not path_or_template.is_file():
                raise ValueError("template is not path to a file or a dictionary")
            with path_or_template.open() as f:
                template = yaml.load(f, Loader=SafeLoader)

        _validate_template(template)
        if name is None:
            name = str(template["name"])

        # Add the template name as an optional key to the renderer schema. This is
        # so that the dictionary passed to the `Renderer.from_dict()` method can
        # contain names of registered templates. These templates are not known when the
        # renderer schema is created.
        # However, this schema is lax because the kwds just has to be an object. Keys
        # and values in kwds are validated in the renderer.
        key = f"template_{name.replace(' ', '_')}"
        _RENDERER_SCHEMA["definitions"][key] = {
            "required": ["name", "kwds"],
            "properties": {"name": {"enum": [name]}, "kwds": {"type": "object"}},
            "additionalProperties": False,
        }
        # Do not add template to `instructions` properties if it has already been added.
        template_ref = {"$ref": f"#/definitions/{key}"}
        oneof = _RENDERER_SCHEMA["properties"]["instructions"]["items"]["oneOf"]
        if template_ref not in oneof:
            oneof.append(template_ref)

        # Add template to registry.
        # TODO: should we log a message if overwriting a key-value pair?
        cls._templates[name.lower()] = template
        return template

    @classmethod
    def get(cls, name: str) -> TemplateType:
        """Return a Template object from the registry given a template name.

        Parameters
        ----------
        name : str
            The name of the registered template.

        If the template is not found, perhaps it was not added to the registry using
        `register`.
        """
        name = name.lower()
        try:
            return cls._templates[name]
        except KeyError:
            known = "', '".join(cls._templates.keys())
            raise TemplateNotFound(
                f"Unknown template '{name}'. Registered templates are '{known}'."
            )

    @classmethod
    def keys(cls) -> ty.KeysView[str]:
        """Return names of registered templates."""
        return cls._templates.keys()

    @classmethod
    def items(cls) -> ty.ItemsView[str, TemplateType]:
        return cls._templates.items()


register_template = _TemplateRegistry.register
registered_templates = _TemplateRegistry.keys
registered_templates_items = _TemplateRegistry.items
get_template = _TemplateRegistry.get



import pytest

from neurodocker.reproenv.exceptions import RendererError
from neurodocker.reproenv.renderers import DockerRenderer
from neurodocker.reproenv.template import Template
from neurodocker.reproenv.tests.utils import prune_rendered


def test_docker_renderer_add_template():
    r = DockerRenderer("apt")

    d = {
        "name": "foobar",
        "binaries": {
            "urls": {"1.0.0": "foobar"},
            "env": {"foo": "bar"},
            "instructions": "{{self.install_dependencies()}}\necho hello\necho world",
            "arguments": {"required": [], "optional": {}},
            "dependencies": {"apt": ["curl"], "debs": [], "yum": ["python"]},
        },
    }

    # Not a Template type.
    with pytest.raises(
        RendererError, match="template must be an instance of 'Template' but got"
    ):
        r.add_template(d, method="binaries")

    # Invalid method
    with pytest.raises(
        RendererError, match="method must be 'binaries', 'source' but got 'fakemethod"
    ):
        r.add_template(Template(d), method="fakemethod")

    # Test apt.
    r.add_template(Template(d), method="binaries")
    assert len(r._parts) == 2
    assert r._parts[0] == 'ENV foo="bar"'
    assert (
        r._parts[1]
        == """RUN apt-get update -qq \\
    && apt-get install -y -q --no-install-recommends \\
           curl \\
    && rm -rf /var/lib/apt/lists/* \\
    && echo hello \\
    && echo world"""
    )

    # Test yum.
    r = DockerRenderer("yum")
    r.add_template(Template(d), method="binaries")
    assert len(r._parts) == 2
    assert r._parts[0] == 'ENV foo="bar"'
    assert (
        r._parts[1]
        == """RUN yum install -y -q \\
           python \\
    && yum clean all \\
    && rm -rf /var/cache/yum/* \\
    && echo hello \\
    && echo world"""
    )

    # Test required arguments.
    d = {
        "name": "foobar",
        "binaries": {
            "urls": {"1.0.0": "foobar"},
            "env": {"foo": "bar"},
            "instructions": (
                "{{self.install_dependencies()}}\necho hello {{ self.name }}"
            ),
            "arguments": {"required": ["name"], "optional": {}},
            "dependencies": {"apt": ["curl"], "debs": [], "yum": ["python"]},
        },
    }
    r = DockerRenderer("apt")
    r.add_template(Template(d, binaries_kwds=dict(name="Bjork")), method="binaries")
    rendered = str(r)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """ENV foo="bar"
RUN apt-get update -qq \\
    && apt-get install -y -q --no-install-recommends \\
           curl \\
    && rm -rf /var/lib/apt/lists/* \\
    && echo hello Bjork"""
    )

    d = {
        "name": "foobar",
        "binaries": {
            "urls": {"1.0.0": "foobar"},
            "env": {"foo": "bar"},
            "instructions": (
                "{{self.install_dependencies()}}\necho hello {{ self.name }}"
            ),
            "arguments": {"required": [], "optional": {"name": "foo"}},
            "dependencies": {"apt": ["curl"], "debs": [], "yum": ["python"]},
        },
    }

    r = DockerRenderer("apt")
    r.add_template(Template(d), method="binaries")
    rendered = str(r)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """ENV foo="bar"
RUN apt-get update -qq \\
    && apt-get install -y -q --no-install-recommends \\
           curl \\
    && rm -rf /var/lib/apt/lists/* \\
    && echo hello foo"""
    )


def test_docker_render_from_instance_methods():
    d = DockerRenderer("apt")

    d.from_("alpine")
    assert prune_rendered(str(d)).strip() == "FROM alpine"

    d = DockerRenderer("apt")
    d.from_("alpine", as_="builder")
    assert prune_rendered(str(d)).strip() == "FROM alpine AS builder"

    d = DockerRenderer("apt")
    d.from_("alpine", as_="builder")
    d.arg("FOO")
    assert prune_rendered(str(d)).strip() == "FROM alpine AS builder\nARG FOO"

    d = DockerRenderer("apt")
    d.from_("alpine", as_="builder")
    d.arg("FOO")
    d.copy(
        ["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/", from_="builder", chown="neuro"
    )
    rendered = str(d)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
FROM alpine AS builder
ARG FOO
COPY --from=builder --chown=neuro ["foo/bar/baz.txt", \\
      "foo/baz/cat.txt", \\
      "/opt/"]"""
    )

    d = DockerRenderer("apt")
    d.from_("alpine", as_="builder")
    d.arg("FOO")
    d.copy(
        ["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/", from_="builder", chown="neuro"
    )
    d.env(PATH="$PATH:/opt/foo/bin")
    rendered = str(d)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
FROM alpine AS builder
ARG FOO
COPY --from=builder --chown=neuro ["foo/bar/baz.txt", \\
      "foo/baz/cat.txt", \\
      "/opt/"]
ENV PATH="$PATH:/opt/foo/bin\""""
    )

    d = DockerRenderer("apt")
    d.from_("alpine", as_="builder")
    d.arg("FOO")
    d.copy(
        ["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/", from_="builder", chown="neuro"
    )
    d.env(PATH="$PATH:/opt/foo/bin")
    d.label(ORG="myorg")
    d.labels({'org.test.label': 'another label'})
    rendered = str(d)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
FROM alpine AS builder
ARG FOO
COPY --from=builder --chown=neuro ["foo/bar/baz.txt", \\
      "foo/baz/cat.txt", \\
      "/opt/"]
ENV PATH="$PATH:/opt/foo/bin"
LABEL ORG="myorg"
LABEL org.test.label="another label\""""
    )

    d = DockerRenderer("apt")
    d.from_("alpine", as_="builder")
    d.arg("FOO")
    d.copy(
        ["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/", from_="builder", chown="neuro"
    )
    d.env(PATH="$PATH:/opt/foo/bin")
    d.label(ORG="myorg")
    d.run("echo foobar")
    rendered = str(d)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
FROM alpine AS builder
ARG FOO
COPY --from=builder --chown=neuro ["foo/bar/baz.txt", \\
      "foo/baz/cat.txt", \\
      "/opt/"]
ENV PATH="$PATH:/opt/foo/bin"
LABEL ORG="myorg"
RUN echo foobar"""
    )

    d = DockerRenderer("apt")
    d.from_("alpine", as_="builder")
    d.arg("FOO")
    d.copy(
        ["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/", from_="builder", chown="neuro"
    )
    d.env(PATH="$PATH:/opt/foo/bin")
    d.label(ORG="myorg")
    d.run("echo foobar")
    d.user("nonroot")
    rendered = str(d)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
FROM alpine AS builder
ARG FOO
COPY --from=builder --chown=neuro ["foo/bar/baz.txt", \\
      "foo/baz/cat.txt", \\
      "/opt/"]
ENV PATH="$PATH:/opt/foo/bin"
LABEL ORG="myorg"
RUN echo foobar
RUN test "$(getent passwd nonroot)" \\
    || useradd --no-user-group --create-home --shell /bin/bash nonroot
USER nonroot"""
    )

    d = DockerRenderer("apt", users={"root", "nonroot"})
    d.from_("alpine", as_="builder")
    d.arg("FOO")
    d.copy(
        ["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/", from_="builder", chown="neuro"
    )
    d.env(PATH="$PATH:/opt/foo/bin")
    d.label(ORG="myorg")
    d.run("echo foobar")
    d.user("nonroot")
    d.workdir("/opt/foobar")
    rendered = str(d)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
FROM alpine AS builder
ARG FOO
COPY --from=builder --chown=neuro ["foo/bar/baz.txt", \\
      "foo/baz/cat.txt", \\
      "/opt/"]
ENV PATH="$PATH:/opt/foo/bin"
LABEL ORG="myorg"
RUN echo foobar
USER nonroot
WORKDIR /opt/foobar"""
    )

    d = DockerRenderer("apt", users={"root", "nonroot"})
    d.from_("alpine", as_="builder")
    d.arg("FOO")
    d.copy(
        ["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/", from_="builder", chown="neuro"
    )
    d.env(PATH="$PATH:/opt/foo/bin")
    d.label(ORG="myorg")
    d.run("echo foobar")
    d.user("nonroot")
    d.workdir("/opt/foobar")
    d.run_bash("source activate")
    rendered = str(d)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
FROM alpine AS builder
ARG FOO
COPY --from=builder --chown=neuro ["foo/bar/baz.txt", \\
      "foo/baz/cat.txt", \\
      "/opt/"]
ENV PATH="$PATH:/opt/foo/bin"
LABEL ORG="myorg"
RUN echo foobar
USER nonroot
WORKDIR /opt/foobar
RUN bash -c 'source activate'"""
    )

    d = DockerRenderer("apt")
    d.from_("debian:buster-slim")
    d.entrypoint(["echo", "foo bar"])
    rendered = str(d)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
FROM debian:buster-slim
ENTRYPOINT ["echo", "foo bar"]"""
    )



"""Minify a Docker container.

This program removes all files under specified directories in the container _not_ used
by the given command(s).
"""

# TODO: consider implementing custom types for Docker container and paths within a
# Docker container.

import io
import logging
from pathlib import Path
import tarfile
import typing as ty

import click

try:
    import docker
except ImportError:
    raise ImportError(
        "The `docker` Python package is required for minification functions."
    )

try:
    client = docker.from_env()
    if not client.ping():
        raise RuntimeError("Could not communicate with the Docker Engine.")
except docker.errors.DockerException:
    raise RuntimeError("Could not create a Docker client.")

logger = logging.getLogger(__name__)

_trace_script = Path(__file__).parent / "_trace.sh"
_prune_script = Path(__file__).parent / "_prune.py"


def copy_file_to_container(
    container: ty.Union[str, docker.models.containers.Container],
    src: ty.Union[str, Path],
    dest: ty.Union[str, Path],
) -> bool:
    """Copy `local_filepath` into `container`:`container_path`.

    Parameters
    ----------
    container : str or `docker.models.containers.Container` instance
        Container to which file is copied.
    src : str or `pathlib.Path` instance
        Filepath on the host.
    dest : str or `pathlib.Path` instance
        Directory inside container. Original filename is preserved.

    Returns
    -------
    success : bool
        True if copy was a success. False otherwise.
    """
    src = Path(src)
    dest = Path(dest)
    if not isinstance(container, docker.models.containers.Container):
        container = client.containers.get(container)
    # https://gist.github.com/zbyte64/6800eae10ce082bb78f0b7a2cca5cbc2
    with io.BytesIO() as tar_stream:
        with tarfile.TarFile(fileobj=tar_stream, mode="w") as tar:
            filename = src.name
            tar.add(src, arcname=filename, recursive=False)
        tar_stream.seek(0)
        return container.put_archive(str(dest), tar_stream)


def _get_mounts(container: docker.models.containers.Container) -> dict:
    # [
    #     {
    #         "Type": "bind",
    #         "Source": "/path/to/source",
    #         "Destination": "/destination",
    #         "Mode": "ro",
    #         "RW": False,
    #         "Propagation": "rprivate",
    #     }
    # ]
    return client.api.inspect_container(container)["Mounts"]


@click.command()
@click.option(
    "-c", "--container", required=True, help="ID or name of running Docker container"
)
@click.option(
    "-d",
    "--dir",
    "directories_to_prune",
    required=True,
    multiple=True,
    help="Directories in container to prune. Data will be lost in these directories",
)
@click.option("--yes", is_flag=True, help="Reply yes to all prompts.")
@click.argument("command", nargs=-1, required=True)
def minify(
    container: ty.Union[str, docker.models.containers.Container],
    directories_to_prune: ty.Tuple[str],
    yes: bool,
    command: ty.Tuple[str],
) -> None:
    """Minify a container.

    Trace COMMAND... in the container, and remove all files in `--dirs-to-prune` that
    were not used by the commands.

    \b
    Examples
    --------
    docker run --rm -itd --name to-minify python:3.9-slim bash
    neurodocker minify \\
        --container to-minify \\
        --dir /usr/local \\
        "python -c 'a = 1 + 1; print(a)'"
    """
    container = client.containers.get(container)
    container = ty.cast(docker.models.containers.Container, container)

    cmds = " ".join(f'"{c}"' for c in command)

    # Copy the trace.sh file into the container and run it.
    copy_file_to_container(container, _trace_script, "/tmp/")
    trace_cmd = f"bash /tmp/_trace.sh {cmds}"
    logger.info(f"running command within container {container.id}: {trace_cmd}")

    # Run container. We need to use the lower-level docker-py API to have access to the
    # exec_id. Using the exec_id, we can test for the exec's exit code with each
    # iteration.
    exec_dict: dict = container.client.api.exec_create(container.id, cmd=trace_cmd)
    exec_id: str = exec_dict["Id"]
    log_gen: ty.Generator[bytes, None, None] = container.client.api.exec_start(
        exec_id, stream=True
    )
    for log in log_gen:
        log_str = log.decode().strip()
        click.secho(log_str, fg="yellow")
    exit_code: int = client.api.exec_inspect(exec_id)["ExitCode"]
    if exit_code != 0:
        raise RuntimeError("error in container")

    # Get files to prune.
    copy_file_to_container(container, _prune_script, "/tmp/")
    ret: int
    result: bytes
    ret, result = container.exec_run(
        "/tmp/reprozip-miniconda/bin/python /tmp/_prune.py"
        " --config-file /tmp/neurodocker-reprozip-trace/config.yml"
        " --dirs-to-prune {}".format(" ".join(map(str, directories_to_prune))).split()
    )
    if ret != 0:
        raise RuntimeError(f"Failed: {result.decode().strip()}")

    ret, result = container.exec_run(
        ["cat", "/tmp/neurodocker-reprozip-trace/TO_DELETE.list"]
    )
    if ret != 0:
        raise RuntimeError(f"Error: {result.decode().strip()}")

    files_to_remove = [Path(p) for p in result.decode().strip().splitlines()]
    if not len(files_to_remove):
        print("No files to remove. Quitting.")
        return

    # Check if any files to be removed are in mounted directories.
    mounts = _get_mounts(container.id)
    if mounts:
        for m in mounts:
            for p in files_to_remove:
                if Path(m["Destination"]) in Path(p).parents:
                    click.get_current_context().fail(
                        "Attempting to remove files in a mounted directory.  Directory"
                        f" in the container '{m['Destination']}' is host directory"
                        f" '{m['Source']}'. Remove this mounted directory from the"
                        " minification command and rerun."
                    )

    click.secho("\nWARNING!!! THE FOLLOWING FILES WILL BE REMOVED:", fg="yellow")
    click.echo()
    click.echo("    " + "\n    ".join(sorted(map(str, files_to_remove))))
    click.secho(
        "\nWARNING: PROCEEDING MAY RESULT IN IRREVERSIBLE DATA LOSS, FOR EXAMPLE"
        " IF ATTEMPTING TO REMOVE FILES IN DIRECTORIES MOUNTED FROM THE HOST."
        " PROCEED WITH EXTREME CAUTION! NEURODOCKER ASSUMES NO RESPONSIBILITY"
        " FOR DATA LOSS. ALL OF THE FILES LISTED ABOVE WILL BE REMOVED.",
        fg="yellow",
    )
    if yes:
        click.secho("Proceeding because --yes was provided", fg="green")
    else:
        click.confirm("Proceed?", abort=True)
    click.echo("Removing files ...")
    ret, result = container.exec_run(
        'xargs -d "\n" -a /tmp/neurodocker-reprozip-trace/TO_DELETE.list rm -f'
    )
    if ret:
        raise RuntimeError(f"Error: {result.decode().split()}")

    ret, result = container.exec_run(
        "rm -rf /tmp/neurodocker-reprozip-trace /tmp/reprozip-miniconda"
        " /tmp/_trace.sh /tmp/_prune.py"
    )
    if ret:
        raise RuntimeError(f"Error: {result.decode().split()}")

    click.secho("\n\nFinished removing files.", fg="green")
    click.echo("Next, create a new Docker image with the minified container:")
    click.echo(f"\n    docker export {container.name} | docker import - imagename\n")



import contextlib
import getpass
import os
from pathlib import Path
import subprocess
import typing as ty
import uuid

import pytest

# TODO: do not use the docker-py package. It is better to be consistent and use
# subprocess for Docker and Singularity. In both cases, we are only
# TODO: only require docker-py for container minification, as it is more useful there
# to get logs as the container is building.
try:
    import docker
except ImportError:
    docker = None


def _docker_available():
    """Return `True` if docker-py is installed or docker engine is available."""
    if docker is None:
        return False
    client = docker.from_env()
    try:
        return client.ping()  # bool, unless engine is unresponsive (eg not installed)
    except docker.errors.APIError:
        return False


def _singularity_available():
    try:
        process = subprocess.run(["singularity", "help"])
        return process.returncode == 0
    except FileNotFoundError:
        return False


# See https://docs.pytest.org/en/stable/skipping.html#id1 for skipif.

skip_if_no_docker = pytest.mark.skipif(
    not _docker_available(), reason="docker not available"
)

skip_if_no_singularity = pytest.mark.skipif(
    not _singularity_available(), reason="singularity not available"
)


@contextlib.contextmanager
def build_docker_image(context: Path, remove=False) -> ty.Generator[str, None, None]:
    """Context manager that builds a Docker image and removes it on exit.

    The argument `remove` is `False` by default because we clean up all images at the
    end of each pytest session.

    Yields
    ------
    str
        Docker image tag.
    """
    df = context / "Dockerfile"
    if not df.exists():
        raise FileNotFoundError(f"Dockerfile not found: {df}")
    tag = "reproenv-pytest-" + uuid.uuid4().hex
    cmd: ty.List[str] = ["docker", "build", "--tag", tag, str(context)]
    try:
        _ = subprocess.check_output(cmd, cwd=context)
        yield tag
    finally:
        # There is no container to remove. That should be done by the tests.
        # Containers made from this image _need_ to be removed for this image to be
        # removed.
        # Do not force image removal, because perhaps others depend on the image.
        # TODO: we do not make use of Docker's caching if we remove...
        if remove:
            subprocess.run(["docker", "image", "rm", tag])
            subprocess.run(["docker", "builder", "prune", "--force"])


@contextlib.contextmanager
def build_singularity_image(
    context: Path, remove=True
) -> ty.Generator[str, None, None]:
    """Context manager that builds a Singularity image and removes it on exit.

    If `sudo singularity` is not available, the full path to `singularity` can be set
    with the environment variable `REPROENV_SINGULARITY_PROGRAM`.

    Yields
    ------
    Path
        Path to Singularity image.
    """
    recipe = context / "Singularity"
    if not recipe.exists():
        raise FileNotFoundError(f"Singularity recipe not found: {recipe}")
    sif = context / f"reproenv-pytest-{uuid.uuid4().hex}.sif"
    # Set singularity cache to /dev/shm
    user = getpass.getuser()
    cachedir = Path("/") / "dev" / "shm" / user / "singularity"
    singularity = os.environ.get("REPROENV_SINGULARITY_PROGRAM", "singularity")
    cmd: ty.List[str] = [
        "sudo",
        f"SINGULARITY_CACHEDIR={cachedir}",
        singularity,
        "build",
        str(sif),
        str(recipe),
    ]
    try:
        _ = subprocess.check_output(cmd, cwd=context)
        yield str(sif)
    finally:
        if remove:
            try:
                sif.unlink()
            except FileNotFoundError:
                pass


def run_docker_image(
    img: str, args: ty.List[str] = None, entrypoint: ty.List[str] = None
):
    """Wrapper for `docker run`.

    Returns
    -------
    str, str
        Stdout and stderr of the process.
    """
    cmd = ["docker", "run", "--rm", "-i"]
    if entrypoint is not None:
        cmd.extend(["--entrypoint", *entrypoint])
    cmd.append(img)
    if args is not None:
        cmd.extend(args)
    print(" ".join(cmd))
    process = subprocess.run(cmd, capture_output=True, check=True)
    stdout = process.stdout.decode().strip()
    stderr = process.stderr.decode().strip()
    return stdout, stderr


def run_singularity_image(
    img: str, args: ty.List[str] = None, entrypoint: ty.List[str] = None
):
    """Wrapper for `singularity run` or `singularity exec`.

    Returns
    -------
    str, str
        Stdout and stderr of the process.
    """
    scmd = "run" if entrypoint is None else "exec"
    # sudo not required
    cmd: ty.List[str] = ["singularity", scmd, "--cleanenv", img]
    if entrypoint is not None:
        cmd.extend(entrypoint)
    if args is not None:
        cmd.extend(args)
    process = subprocess.run(cmd, capture_output=True, check=True)
    stdout = process.stdout.decode().strip()
    stderr = process.stderr.decode().strip()
    return stdout, stderr


def get_build_and_run_fns(cmd: str):
    cmd = cmd.lower()
    if cmd == "docker":
        return build_docker_image, run_docker_image
    elif cmd == "singularity":
        return build_singularity_image, run_singularity_image
    else:
        raise KeyError(f"unknown cmd: {cmd}")


def prune_rendered(r: str) -> str:
    # Remove portion that saves JSON to file.
    lines = r.splitlines()
    start_bad = lines.index("# Save specification to JSON.")
    end_bad = lines.index("# End saving to specification to JSON.")
    new_lines = lines[: start_bad - 1] + lines[end_bad + 1 :]
    r = "\n".join(new_lines)

    # Remove comments.
    # This part has to happen _after_ removing the JSON, because
    # JSON is removed based on a certain comment.
    r = "\n".join(line for line in r.splitlines() if not line.startswith("#"))

    # TODO: should we remove empty lines?

    return r



"""Neurodocker generates containers for neuroimaging.

It includes ReproEnv, which is an extensible, generic container generator.
"""

from pathlib import Path

from neurodocker._version import get_versions
from neurodocker import reproenv  # noqa: F401

__version__ = get_versions()["version"]
del get_versions

# Register neurodocker templates
# TODO: remove registration from the `generate` cli. otherwise we register twice.
for template_path in (Path(__file__).parent / "templates").glob("*.yaml"):
    reproenv.register_template(template_path)

del template_path, Path



"""Container specification renderers."""

from __future__ import annotations

import functools
import inspect
import json
import os
import pathlib
import types
import typing as ty

import jinja2

from neurodocker.reproenv.exceptions import RendererError
from neurodocker.reproenv.exceptions import TemplateError
from neurodocker.reproenv.state import _TemplateRegistry
from neurodocker.reproenv.state import _validate_renderer
from neurodocker.reproenv.template import _BaseInstallationTemplate
from neurodocker.reproenv.template import Template
from neurodocker.reproenv.types import _SingularityHeaderType
from neurodocker.reproenv.types import REPROENV_SPEC_FILE_IN_CONTAINER
from neurodocker.reproenv.types import allowed_pkg_managers
from neurodocker.reproenv.types import allowed_installation_methods
from neurodocker.reproenv.types import installation_methods_type
from neurodocker.reproenv.types import pkg_managers_type

# All jinja2 templates are instantiated from this environment object. It is
# configured to dislike undefined attributes. For example, if a template is
# created with the string '{{ foo.bar }}' and 'foo' does not have a 'bar'
# attribute, an error will be thrown when the jinja template is instantiated.
_jinja_env = jinja2.Environment(undefined=jinja2.StrictUndefined)

# Add globals to the jinja environment. These functions can be called from within a
# template.


def _raise_helper(msg: str) -> ty.NoReturn:
    raise RendererError(msg)


_jinja_env.globals["raise"] = _raise_helper

# TODO: add a flag that avoids buggy behavior when basing a new container on
# one created with ReproEnv.

PathType = ty.Union[str, pathlib.Path, os.PathLike]


def _render_string_from_template(
    source: str, template: _BaseInstallationTemplate
) -> str:
    """Take a string from a template and render """
    # TODO: we could use a while loop or recursive function to render the template until
    # there are no jinja-specific things. At this point, we support one level of
    # nesting.
    n_renders = 0
    max_renders = 20

    err = (
        "A template included in this renderer raised an error. Please check the"
        " template definition. A required argument might not be included in the"
        " required arguments part of the template. Variables in the template should"
        " start with `self.`."
    )

    # Render the string again. This is sometimes necessary because some defaults in the
    # template are rendered as {{ self.X }}. These defaults need to be rendered again.

    while (
        _jinja_env.variable_start_string in source
        and _jinja_env.variable_end_string in source
    ):
        source = source.replace("self.", "template.")
        tmpl = _jinja_env.from_string(source)
        try:
            source = tmpl.render(template=template)
        except jinja2.exceptions.UndefinedError as e:
            raise RendererError(err) from e
        n_renders += 1

        if n_renders > max_renders:
            raise RendererError(
                f"reached maximum rendering iterations ({max_renders}). Templates"
                f" should not nest variables more than {max_renders} times."
            )
    return source


def _log_instruction(func: ty.Callable):
    """Decorator that logs instructions passed to a Renderer.

    This adds the logs to the `_instructions` attribute of the Renderer instance.
    """

    @functools.wraps(func)
    def with_logging(self, *args, **kwds):
        if not hasattr(self, "_instructions"):
            raise ValueError(
                "This wrapper should only be applied to Renderer instances."
            )
        # This gets all arguments, even if their value is None.
        # We remove key-value pairs where value is None, because the schema does not
        # allow for None values.
        # TODO: consider this further...
        sig = inspect.signature(func)

        # We could apply defaults with `bargs.apply_defaults()` but we do not because
        # many defaults are None and the renderer schema does not support null values.
        bargs = sig.bind(self, *args, **kwds)
        # self is not an argument in the spec, but it is present because these are
        # instance methods.
        del bargs.arguments["self"]

        # If a function takes **kwds, save those without wrapping them in another dict.
        # Assume that **kwds arguments are _always_ kwds (eg, not kwargs).
        # TODO: generalize this to work on any VAR_KEYWORD parameter.
        kwds_param = sig.parameters.get("kwds")
        if kwds_param is not None:
            if kwds_param.kind == kwds_param.VAR_KEYWORD:
                bargs_kwds = bargs.arguments.pop("kwds")
                if bargs_kwds is not None:
                    bargs.arguments.update(bargs_kwds)

        d = {"name": func.__name__, "kwds": dict(bargs.arguments)}

        self._instructions["instructions"].append(d)
        return func(self, *args, **kwds)

    return with_logging


class _Renderer:
    def __init__(
        self, pkg_manager: pkg_managers_type, users: ty.Optional[ty.Set[str]] = None
    ) -> None:
        if pkg_manager not in allowed_pkg_managers:
            raise RendererError(
                "Unknown package manager '{}'. Allowed package managers are"
                " '{}'.".format(pkg_manager, "', '".join(allowed_pkg_managers))
            )

        self.pkg_manager = pkg_manager
        self._users = {"root"} if users is None else users
        # This keeps track of the current user. This is useful when saving the JSON
        # specification to JSON, because if we are not root, we can change to root,
        # write the file, and return to whichever user we were.
        self._current_user = "root"
        self._instructions: ty.Mapping = {
            "pkg_manager": self.pkg_manager,
            "existing_users": list(self._users),
            "instructions": [],
        }

        # Strings (comments) that indicate the beginning and end of saving the JSON
        # spec to a file. This helps us in testing because sometimes we don't care
        # about the JSON but we do care about everything else. We can remove the
        # content between these two strings.
        self._json_save_start = "# Save specification to JSON."
        self._json_save_end = "# End saving to specification to JSON."

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (_Renderer, str)):
            raise NotImplementedError()

        def rm_empty_lines(s):
            return "\n".join(
                j
                for j in str(s).splitlines()
                if j.strip() and not j.strip().startswith("#")
            )

        # Empty lines and commented lines do not affect container definitions.
        return rm_empty_lines(self) == rm_empty_lines(other)

    def __str__(self) -> str:
        masthead = "# Generated by Neurodocker and Reproenv."
        image_spec = self.render()
        return f"{masthead}\n\n{image_spec}"

    @property
    def users(self) -> ty.Set[str]:
        return self._users

    @classmethod
    def from_dict(cls, d: ty.Mapping) -> _Renderer:
        """Instantiate a new renderer from a dictionary of instructions."""
        # raise error if invalid
        _validate_renderer(d)

        pkg_manager = d["pkg_manager"]
        users = d.get("existing_users", None)

        # create new renderer object
        renderer = cls(pkg_manager=pkg_manager, users=users)

        for mapping in d["instructions"]:
            method_or_template = mapping["name"]
            kwds = mapping["kwds"]
            this_instance_method = getattr(renderer, method_or_template, None)
            # Method exists and is something like 'copy', 'env', 'run', etc.
            if this_instance_method is not None:
                try:
                    this_instance_method(**kwds)
                except Exception as e:
                    raise RendererError(
                        f"Error on step '{method_or_template}'. Please see the"
                        " traceback above for details."
                    ) from e
            # This is actually a template.
            else:
                try:
                    renderer.add_registered_template(method_or_template, **kwds)
                except TemplateError as e:
                    raise RendererError(
                        f"Error on template '{method_or_template}'. Please see above"
                        " for more information."
                    ) from e
        return renderer

    def render(self) -> str:
        """Return a rendered string of the container specification.

        This method should take care of combining all of the instructions in a way that
        makes sense for the container type. For a Dockerfile, it might stack the
        sequence of instructions into a string. For a Singularity recipe, it might
        organize the instructions into their respective locations, like %post, %files.

        This method is used by Renderer.__str__ to create the full container spec.
        """
        raise NotImplementedError("subclasses must implement .render()")

    def add_template(
        self, template: Template, method: installation_methods_type
    ) -> _Renderer:
        """Add a template to the renderer.

        Parameters
        ----------
        template : Template
            The template to add. To reference templates by name, use
            `.add_registered_template`.
        method : str
            The method to use to install the software described in the template.
        """

        if not isinstance(template, Template):
            raise RendererError(
                "template must be an instance of 'Template' but got"
                f" '{type(template)}'."
            )
        if method not in allowed_installation_methods:
            raise RendererError(
                "method must be '{}' but got '{}'.".format(
                    "', '".join(sorted(allowed_installation_methods)), method
                )
            )

        template_method: _BaseInstallationTemplate = getattr(template, method)
        if template_method is None:
            raise RendererError(f"template does not have entry for: '{method}'")
        # Validate kwds passed by user to template, and raise an exception if any are
        # invalid.
        template_method.validate_kwds()

        # TODO: print a message if the template has a nonempty `alert` property.
        # If we print to stdout, however, we can cause problems if the user is piping
        # the output to a file or directly to a container build command.

        # If we keep the `self.VAR` syntax of the template, then we need to pass
        # `self=template_method` to the renderer function. But that function is an
        # instance method, so passing `self` will override the `self` argument.
        # To get around this, we replace `self.` with something that is not an
        # argument to the renderer function.

        # Add environment (render any jinja templates).
        if template_method.env:
            d: ty.Mapping[str, str] = {
                _render_string_from_template(
                    k, template_method
                ): _render_string_from_template(v, template_method)
                for k, v in template_method.env.items()
            }
            self.env(**d)

        # Patch the `template_method.install_dependencies` instance method so it can be
        # used (ie rendered) in a template and have access to the pkg_manager requested.
        def install_patch(
            inner_self: _BaseInstallationTemplate, pkgs: ty.List[str], opts: str = None
        ) -> str:
            return _install(pkgs=pkgs, pkg_manager=self.pkg_manager)

        # mypy complains when we try to patch a class, so we do it behind its back with
        # setattr. See https://github.com/python/mypy/issues/2427
        setattr(
            template_method, "install", types.MethodType(install_patch, template_method)
        )

        # Set pkg_manager onto the template.
        setattr(template_method, "pkg_manager", self.pkg_manager)

        # Patch the `template_method.install_dependencies` instance method so it can be
        # used (ie rendered) in a template and have access to the pkg_manager requested.
        def install_dependencies_patch(
            inner_self: _BaseInstallationTemplate, opts: str = None
        ) -> str:
            # TODO: test that template with empty dependencies (apt: []) does not render
            # any installation of dependencies.
            cmd = ""
            pkgs = inner_self.dependencies(pkg_manager=self.pkg_manager)
            if pkgs:
                cmd += _install(pkgs=pkgs, pkg_manager=self.pkg_manager, opts=opts)
            if self.pkg_manager == "apt":
                debs = inner_self.dependencies("debs")
                if debs:
                    cmd += "\n" + _apt_install_debs(debs)
            return cmd

        # mypy complains when we try to patch a class, so we do it behind its back with
        # setattr. See https://github.com/python/mypy/issues/2427
        setattr(
            template_method,
            "install_dependencies",
            types.MethodType(install_dependencies_patch, template_method),
        )

        # Add installation instructions (render any jinja templates).
        if template_method.instructions:
            command = _render_string_from_template(
                template_method.instructions, template_method
            )
            # TODO: raise exception here or skip the run instruction?
            if not command.strip():
                raise RendererError(f"empty rendered instructions in {template.name}")
            self.run(command)

        return self

    def add_registered_template(
        self, name: str, method: installation_methods_type = None, **kwds
    ) -> _Renderer:

        # Template was validated at registration time.
        template_dict = _TemplateRegistry.get(name)

        # By default, prefer 'binaries', but use 'source' if 'binaries' is not defined.
        # TODO: should we require user to provide method?
        if method is None:
            method = "binaries" if "binaries" in template_dict else "source"
        if method not in template_dict:
            raise RendererError(
                f"Installation method '{method}' not defined for template '{name}'."
                " Options are '{}'.".format("', '".join(template_dict.keys()))
            )

        binaries_kwds = source_kwds = None
        if method == "binaries":
            binaries_kwds = kwds
        elif method == "source":
            source_kwds = kwds

        template = Template(
            template=template_dict, binaries_kwds=binaries_kwds, source_kwds=source_kwds
        )

        self.add_template(template=template, method=method)
        return self

    def arg(self, key: str, value: str = None):
        raise NotImplementedError()

    def copy(
        self,
        source: ty.Union[PathType, ty.List[PathType]],
        destination: ty.Union[PathType, ty.List[PathType]],
    ) -> _Renderer:
        raise NotImplementedError()

    def env(self, **kwds: str) -> _Renderer:
        raise NotImplementedError()

    def entrypoint(self, args: ty.List[str]) -> _Renderer:
        raise NotImplementedError()

    def from_(self, base_image: str) -> _Renderer:
        raise NotImplementedError()

    def install(self, pkgs: ty.List[str], opts: str = None) -> _Renderer:
        raise NotImplementedError()

    def label(self, **kwds: str) -> _Renderer:
        raise NotImplementedError()

    def labels(self, labels_dict: dict) -> _Renderer:
        """Adds a set of labels to the dockerfile from a dict. This permits
        labels that can include special chars (e.g. '.')."""
        self.label(**labels_dict)
        return self

    def run(self, command: str) -> _Renderer:
        raise NotImplementedError()

    def run_bash(self, command: str) -> _Renderer:
        command = f"bash -c '{command}'"
        return self.run(command)

    def user(self, user: str) -> _Renderer:
        raise NotImplementedError()

    def workdir(self, path: PathType) -> _Renderer:
        raise NotImplementedError()

    def to_json(self, **json_kwds) -> str:
        """Return instructions used in this renderer in JSON format.

        json_kwds are keyword arguments passed to `json.dumps`.
        """
        return json.dumps(self._instructions, **json_kwds)

    def _get_instructions(self) -> str:
        """Return string representation of a printf command that writes the renderer
        instructions to a JSON file in the container.
        """
        j = self.to_json(indent=2)
        # Double-escape escaped sequences so that when printf is done with them, they
        # are escaped with a single slash.
        j = j.replace("\\", "\\\\")
        # Same with parentheses.
        j = j.replace("(", "\\(").replace(")", "\\)")
        # Add slash to the end of each line, except the last.
        j = " \\\n".join(j.splitlines())
        # Escape the % characters so printf does not interpret them as delimiters.
        j = j.replace("%", "%%")
        # Escape single quotes with '"'"'
        j = j.replace("'", "'\"'\"'")
        cmd = f"printf '{j}' > {REPROENV_SPEC_FILE_IN_CONTAINER}"
        return cmd


class DockerRenderer(_Renderer):
    def __init__(
        self, pkg_manager: pkg_managers_type, users: ty.Set[str] = None
    ) -> None:
        super().__init__(pkg_manager=pkg_manager, users=users)
        self._parts: ty.List[str] = []

    def render(self) -> str:
        """Return the rendered Dockerfile."""
        s = "\n".join(self._parts)

        # Save specification to JSON.
        s += f"\n\n{self._json_save_start}"
        if self._current_user != "root":
            s += "\nUSER root"
        s += f"\nRUN {self._get_instructions()}"
        if self._current_user != "root":
            s += f"\nUSER {self._current_user}"
        s += f"\n{self._json_save_end}"
        s = s.strip()  # Prune whitespace from beginning and end.
        return s

    @_log_instruction
    def arg(self, key: str, value: str = None) -> DockerRenderer:
        """Add a Dockerfile `ARG` instruction."""
        s = f"ARG {key}" if value is None else f"ARG {key}={value}"
        self._parts.append(s)
        return self

    @_log_instruction
    def copy(
        self,
        source: ty.Union[PathType, ty.List[PathType]],
        destination: PathType,
        from_: str = None,
        chown: str = None,
    ) -> DockerRenderer:
        """Add a Dockerfile `COPY` instruction."""
        if not isinstance(source, (list, tuple)):
            source = [source]
        source.append(destination)
        files = '["{}"]'.format('", \\\n      "'.join(map(str, source)))
        s = "COPY "
        if from_ is not None:
            s += f"--from={from_} "
        if chown is not None:
            s += f"--chown={chown} "
        s += files
        self._parts.append(s)
        return self

    @_log_instruction
    def env(self, **kwds: str) -> DockerRenderer:
        """Add a Dockerfile `ENV` instruction."""
        s = "ENV " + " \\\n    ".join(f'{k}="{v}"' for k, v in kwds.items())
        self._parts.append(s)
        return self

    @_log_instruction
    def entrypoint(self, args: ty.List[str]) -> DockerRenderer:
        s = 'ENTRYPOINT ["{}"]'.format('", "'.join(args))
        self._parts.append(s)
        return self

    @_log_instruction
    def from_(self, base_image: str, as_: str = None) -> DockerRenderer:
        """Add a Dockerfile `FROM` instruction."""
        if as_ is None:
            s = "FROM " + base_image
        else:
            s = f"FROM {base_image} AS {as_}"
        self._parts.append(s)
        return self

    @_log_instruction
    def install(self, pkgs: ty.List[str], opts=None) -> DockerRenderer:
        """Install system packages."""
        command = _install(pkgs, pkg_manager=self.pkg_manager, opts=opts)
        command = _indent_run_instruction(command)
        self.run(command)
        return self

    @_log_instruction
    def label(self, **kwds: str) -> DockerRenderer:
        """Add a Dockerfile `LABEL` instruction."""
        s = "LABEL " + " \\\n      ".join(f'{k}="{v}"' for k, v in kwds.items())
        self._parts.append(s)
        return self

    @_log_instruction
    def run(self, command: str) -> DockerRenderer:
        """Add a Dockerfile `RUN` instruction."""
        # TODO: should the command be quoted?
        # s = shlex.quote(command)
        # if s.startswith("'"):
        #     s = s[1:-1]  # Remove quotes on either end of the string.
        s = command
        s = _indent_run_instruction(f"RUN {s}")
        self._parts.append(s)
        return self

    @_log_instruction
    def user(self, user: str) -> DockerRenderer:
        """Add a Dockerfile `USER` instruction. If the user is not in
        `self.users`, then a `RUN` instruction that creates the user
        will also be added.
        """
        s = ""
        if user not in self._users:
            s = (
                f'RUN test "$(getent passwd {user})" \\\n    || useradd '
                f"--no-user-group --create-home --shell /bin/bash {user}"
            )
            self._parts.append(s)
            self._users.add(user)
        self._parts.append(f"USER {user}")
        self._current_user = user
        return self

    @_log_instruction
    def workdir(self, path: PathType) -> DockerRenderer:
        """Add a Dockerfile `WORKDIR` instruction."""
        self._parts.append("WORKDIR " + str(path))
        return self


class SingularityRenderer(_Renderer):
    def __init__(
        self, pkg_manager: pkg_managers_type, users: ty.Optional[ty.Set[str]] = None
    ) -> None:
        super().__init__(pkg_manager=pkg_manager, users=users)

        self._header: _SingularityHeaderType = {}
        # The '%setup' section is intentionally ommitted.
        self._files: ty.List[str] = []
        self._environment: ty.List[ty.Tuple[str, str]] = []
        self._post: ty.List[str] = []
        self._runscript = ""
        # TODO: is it OK to use a dict here? Labels could be overwritten.
        self._labels: ty.Dict[str, str] = {}

    def render(self) -> str:
        s = ""
        # Create header.
        if self._header:
            s += (
                f"Bootstrap: {self._header['bootstrap']}\nFrom: {self._header['from_']}"
            )

        # Add files.
        if self._files:
            s += "\n\n%files\n"
            s += "\n".join(self._files)

        # Add environment.
        if self._environment:
            s += "\n\n%environment"
            for k, v in self._environment:
                s += f'\nexport {k}="{v}"'

        # Add post.
        # There will always be a post section, because we always want to add the
        # reproenv specification.
        s += "\n\n%post\n"
        # This section might be empty, but that is OK.
        s += "\n\n".join(self._post)
        s += f"\n\n{self._json_save_start}"
        if self._current_user != "root":
            s += "\nsu - root"
        s += f"\n{self._get_instructions()}"
        if self._current_user != "root":
            s += f"\nsu - {self._current_user}"
        s += f"\n{self._json_save_end}"

        # Add runscript.
        if self._runscript:
            s += "\n\n%runscript\n"
            s += self._runscript

        # Add labels.
        if self._labels:
            s += "\n\n%labels"
            for kv in self._labels.items():
                s += "\n" + " ".join(kv)

        return s

    @_log_instruction
    def arg(self, key: str, value: str = None) -> SingularityRenderer:
        # TODO: look into whether singularity has something like ARG, like passing in
        # environment variables.
        s = f"{key}=" if value is None else f"{key}={value}"
        self._post.append(s)
        return self

    @_log_instruction
    def copy(
        self,
        source: ty.Union[PathType, ty.List[PathType]],
        destination: PathType,
    ) -> SingularityRenderer:
        if not isinstance(source, (list, tuple)):
            source = [source]
        files = [f"{src} {destination}" for src in source]
        self._files.extend(files)
        return self

    @_log_instruction
    def env(self, **kwds: str) -> SingularityRenderer:
        # TODO: why does this raise a type error?
        self._environment.extend(kwds.items())  # type: ignore
        return self

    @_log_instruction
    def entrypoint(self, args: ty.List[str]) -> SingularityRenderer:
        self._runscript = " ".join(args)
        return self

    @_log_instruction
    def from_(self, base_image: str) -> SingularityRenderer:
        if "://" not in base_image:
            bootstrap = "docker"
            image = base_image
        elif base_image.startswith("docker://"):
            bootstrap = "docker"
            image = base_image[9:]
        elif base_image.startswith("library://"):
            bootstrap = "library"
            image = base_image[10:]
        else:
            raise RendererError("Unknown singularity bootstrap agent.")

        self._header = {"bootstrap": bootstrap, "from_": image}
        return self

    @_log_instruction
    def install(self, pkgs: ty.List[str], opts=None) -> SingularityRenderer:
        """Install system packages."""
        command = _install(pkgs, pkg_manager=self.pkg_manager, opts=opts)
        self.run(command)
        return self

    @_log_instruction
    def label(self, **kwds: str) -> SingularityRenderer:
        # TODO: why are we getting this error?
        # Argument 1 to "update" of "dict" has incompatible type
        # "Dict[str, Mapping[str, str]]"; expected "Mapping[str, str]"
        self._labels.update(kwds)  # type: ignore
        return self

    @_log_instruction
    def run(self, command: str) -> SingularityRenderer:
        self._post.append(command)
        return self

    @_log_instruction
    def user(self, user: str) -> SingularityRenderer:
        if user not in self._users:
            post = (
                f'test "$(getent passwd {user})" \\\n|| useradd '
                f"--no-user-group --create-home --shell /bin/bash {user}\n"
            )
            self._users.add(user)
            self._post.append(post)
        self._post.append(f"su - {user}")
        self._current_user = user
        return self

    @_log_instruction
    def workdir(self, path: PathType) -> SingularityRenderer:
        self._post.append(f"mkdir -p {path}\ncd {path}")
        return self


def _indent_run_instruction(string: str, indent=4) -> str:
    """Return indented string for Dockerfile `RUN` command."""
    out = []
    lines = string.splitlines()
    for ii, line in enumerate(lines):
        line = line.rstrip()
        is_last_line = ii == len(lines) - 1
        already_cont = line.startswith(("&&", "&", "||", "|", "fi"))
        is_comment = line.startswith("#")
        previous_cont = lines[ii - 1].endswith("\\") or lines[ii - 1].startswith("if")
        if ii:  # do not apply to first line
            if not already_cont and not previous_cont and not is_comment:
                line = "&& " + line
            if not already_cont and previous_cont:
                line = " " * (indent + 3) + line  # indent + len("&& ")
            else:
                line = " " * indent + line
        if not is_last_line and not line.endswith("\\") and not is_comment:
            line += " \\"
        out.append(line)
    return "\n".join(out)


def _install(pkgs: ty.List[str], pkg_manager: str, opts: str = None) -> str:
    if pkg_manager == "apt":
        return _apt_install(pkgs, opts)
    elif pkg_manager == "yum":
        return _yum_install(pkgs, opts)
    # TODO: add debs here?
    else:
        raise RendererError(f"Unknown package manager '{pkg_manager}'.")


def _apt_install(pkgs: ty.List[str], opts: str = None, sort=True) -> str:
    """Return command to install deb packages with `apt-get` (Debian-based distros).

    `opts` are options passed to `yum install`. Default is "-q --no-install-recommends".
    """
    pkgs = sorted(pkgs) if sort else pkgs
    opts = "-q --no-install-recommends" if opts is None else opts
    s = """\
apt-get update -qq
apt-get install -y {opts} \\
    {pkgs}
rm -rf /var/lib/apt/lists/*
""".format(
        opts=opts, pkgs=" \\\n    ".join(pkgs)
    )
    return s.strip()


def _apt_install_debs(urls: ty.List[str], opts: str = None, sort=True) -> str:
    """Return command to install deb packages with `apt-get` (Debian-based distros).

    `opts` are options passed to `yum install`. Default is "-q".
    """

    def install_one(url: str):
        return f"""\
_reproenv_tmppath="$(mktemp -t tmp.XXXXXXXXXX.deb)"
curl -fsSL --retry 5 -o "${{_reproenv_tmppath}}" {url}
apt-get install --yes {opts} "${{_reproenv_tmppath}}"
rm "${{_reproenv_tmppath}}\""""

    urls = sorted(urls) if sort else urls
    opts = "-q" if opts is None else opts

    s = "\n".join(map(install_one, urls))
    s += """
apt-get update -qq
apt-get install --yes --quiet --fix-missing
rm -rf /var/lib/apt/lists/*"""
    return s


def _yum_install(pkgs: ty.List[str], opts: str = None, sort=True) -> str:
    """Return command to install packages with `yum` (CentOS, Fedora).

    `opts` are options passed to `yum install`. Default is "-q".
    """
    pkgs = sorted(pkgs) if sort else pkgs
    opts = "-q" if opts is None else opts

    s = """\
yum install -y {opts} \\
    {pkgs}
yum clean all
rm -rf /var/cache/yum/*
""".format(
        opts=opts, pkgs=" \\\n    ".join(pkgs)
    )
    return s.strip()



# TODO: add more tests for `from_dict` method.

from pathlib import Path
import typing as ty

import pytest

from neurodocker.reproenv.renderers import DockerRenderer
from neurodocker.reproenv.renderers import SingularityRenderer
from neurodocker.reproenv.state import _TemplateRegistry
from neurodocker.reproenv.tests.utils import get_build_and_run_fns
from neurodocker.reproenv.tests.utils import skip_if_no_docker
from neurodocker.reproenv.tests.utils import skip_if_no_singularity
from neurodocker.reproenv.types import installation_methods_type
from neurodocker.reproenv.types import pkg_managers_type

_template_filepath = Path(__file__).parent / "sample-template-jq.yaml"


@pytest.mark.verylong
@pytest.mark.parametrize(
    "cmd",
    [
        pytest.param("docker", marks=skip_if_no_docker),
        pytest.param("singularity", marks=skip_if_no_singularity),
    ],
)
@pytest.mark.parametrize(
    ["pkg_manager", "base_image"], [("apt", "debian:buster-slim"), ("yum", "fedora:33")]
)
@pytest.mark.parametrize(
    ["jq_version", "jq_version_output", "fd_version_startswith"],
    [("1.6", "jq-1.6", "fd")],
)
def test_build_using_renderer_from_dict(
    cmd: str,
    pkg_manager: str,
    base_image: str,
    jq_version: str,
    jq_version_output: str,
    fd_version_startswith: str,
    tmp_path: Path,
):

    _TemplateRegistry._reset()
    _TemplateRegistry.register(_template_filepath)

    d = {
        "pkg_manager": pkg_manager,
        "instructions": [
            {"name": "from_", "kwds": {"base_image": base_image}},
            {"name": "run", "kwds": {"command": "echo hello there"}},
            {"name": "jq", "kwds": {"version": jq_version, "method": "binaries"}},
        ],
    }

    fd_exe = "fdfind" if pkg_manager == "apt" else "fd"

    rcls = DockerRenderer if cmd == "docker" else SingularityRenderer
    specf = "Dockerfile" if cmd == "docker" else "Singularity"

    r = rcls.from_dict(d)
    (tmp_path / specf).write_text(str(r))

    build_fn, run_fn = get_build_and_run_fns(cmd)
    with build_fn(tmp_path) as img:
        stdout, _ = run_fn(img, args=["jq", "--help"])
        assert stdout.startswith("jq - commandline JSON processor")
        stdout, _ = run_fn(img, args=["jq", "--version"])
        assert stdout == jq_version_output
        # Test that deb was installed
        stdout, _ = run_fn(img, args=[fd_exe, "--version"])
        assert stdout.startswith(fd_version_startswith)


@pytest.mark.verylong
@pytest.mark.parametrize(
    "cmd",
    [
        pytest.param("docker", marks=skip_if_no_docker),
        pytest.param("singularity", marks=skip_if_no_singularity),
    ],
)
@pytest.mark.parametrize(
    ["pkg_manager", "base_image"], [("apt", "debian:buster-slim"), ("yum", "fedora:33")]
)
@pytest.mark.parametrize(["method"], [("binaries",), ("source",)])
@pytest.mark.parametrize(
    ["jq_version", "jq_version_output", "fd_version_startswith"],
    [("1.6", "jq-1.6", "fd"), ("1.5", "jq-1.5", "fd")],
)
def test_build_using_renderer_instance_methods(
    cmd: str,
    pkg_manager: str,
    base_image: str,
    method: str,
    jq_version: str,
    jq_version_output: str,
    fd_version_startswith: str,
    tmp_path: Path,
):
    _TemplateRegistry._reset()
    _TemplateRegistry.register(_template_filepath)

    pkg_manager = ty.cast(pkg_managers_type, pkg_manager)
    method = ty.cast(installation_methods_type, method)

    fd_exe = "fdfind" if pkg_manager == "apt" else "fd"

    rcls = DockerRenderer if cmd == "docker" else SingularityRenderer
    r = rcls(pkg_manager=pkg_manager)
    r.from_(base_image)
    r.add_registered_template("jq", method=method, version=jq_version)
    specf = "Dockerfile" if cmd == "docker" else "Singularity"
    (tmp_path / specf).write_text(str(r))

    build_fn, run_fn = get_build_and_run_fns(cmd)
    with build_fn(tmp_path) as img:
        stdout, _ = run_fn(img, args=["jq", "--help"])
        assert stdout.startswith("jq - commandline JSON processor")
        stdout, _ = run_fn(img, args=["jq", "--version"])
        if method == "source" and jq_version == "1.5":
            assert stdout == "jq-"
        else:
            assert stdout == jq_version_output
        # Test that deb was installed
        if method == "binaries":
            stdout, _ = run_fn(img, args=[fd_exe, "--version"])
            assert stdout.startswith(fd_version_startswith)



import pytest

from neurodocker.reproenv.renderers import DockerRenderer
from neurodocker.reproenv.renderers import SingularityRenderer
from neurodocker.reproenv.tests.utils import get_build_and_run_fns
from neurodocker.reproenv.tests.utils import skip_if_no_docker
from neurodocker.reproenv.tests.utils import skip_if_no_singularity


@pytest.mark.parametrize(
    "cmd",
    [
        pytest.param("docker", marks=skip_if_no_docker),
        pytest.param("singularity", marks=skip_if_no_singularity),
    ],
)
def test_build_simple(cmd: str, tmp_path):

    rcls = DockerRenderer if cmd == "docker" else SingularityRenderer

    # Create a Dockerfile.
    r = rcls("apt")
    if isinstance(r, DockerRenderer):
        r.from_("debian:buster-slim", as_="builder")
    else:
        r.from_("debian:buster-slim")
    r.arg("FOO")
    r.copy(["foo.txt", "tst/baz.txt"], "/opt/")
    r.env(PATH="$PATH:/opt/foo/bin")
    r.label(ORG="myorg")
    r.run("echo foobar")
    r.user("nonroot")
    r.workdir("/opt/foobar")
    r.entrypoint(["echo", "hi there"])

    # Create the paths that are copied in the Dockerfile.
    (tmp_path / "tst").mkdir(exist_ok=True)
    (tmp_path / "foo.txt").write_text("")
    (tmp_path / "tst" / "baz.txt").write_text("")

    spec = "Dockerfile" if cmd == "docker" else "Singularity"
    (tmp_path / spec).write_text(str(r))

    build_fn, run_fn = get_build_and_run_fns(cmd)
    with build_fn(tmp_path) as img:
        stdout, _ = run_fn(img)
        assert stdout == "hi there"
        stdout, _ = run_fn(img, entrypoint=["ls"], args=["/opt/"])
        assert set(stdout.splitlines()) == {"baz.txt", "foo.txt", "foobar"}



"""Exceptions for ReproEnv."""


class ReproEnvError(Exception):
    """Base class for all ReproEnv exceptions."""


class RendererError(ReproEnvError):
    pass


class RequirementsError(ReproEnvError):
    pass


class TemplateError(ReproEnvError):
    pass


class TemplateKeywordArgumentError(TemplateError):
    """Invalid keyword argument passed to template, or required argument not
    provided.
    """

    pass


class TemplateNotFound(ReproEnvError):
    pass



import pytest

from neurodocker.reproenv.exceptions import RendererError
from neurodocker.reproenv.renderers import _Renderer
from neurodocker.reproenv.renderers import DockerRenderer
from neurodocker.reproenv.renderers import SingularityRenderer


def test_renderer():
    with pytest.raises(RendererError):
        _Renderer(pkg_manager="foo")

    _Renderer("apt")
    assert _Renderer("yum").users == {"root"}
    assert _Renderer("apt", users={"root", "foo"}).users == {"root", "foo"}


@pytest.mark.parametrize("renderer_cls", [DockerRenderer, SingularityRenderer])
def test_renderer_eq(renderer_cls):
    # r: _Renderer = renderer("apt").from_("my_base_image").run("echo foobar")
    r = renderer_cls("apt").from_("my_base_image").run("echo foobar")
    # print(r)
    assert r == r
    assert r != f"{r}\necho foobar"
    # empty lines do not affect equality
    assert r == f"{r}\n"
    # comments do not count
    assert r == f"{r}\n# a comment"
    # inline comments count
    assert r != f"{r}\necho foo # inline comment"

    r2: _Renderer = renderer_cls("apt").from_("my_base_image").run("echo foobar")
    assert r == r2
    assert r2 == r
    r2.run("echo invalidate this")
    assert r != r2
    assert r2 != r


def test_not_implemented_methods():
    r = _Renderer("yum")
    with pytest.raises(NotImplementedError):
        r.arg(key="a", value="b")
    with pytest.raises(NotImplementedError):
        r.copy("", "")
    with pytest.raises(NotImplementedError):
        r.env(foo="bar")
    with pytest.raises(NotImplementedError):
        r.from_("baseimage")
    with pytest.raises(NotImplementedError):
        r.label(foo="bar")
    with pytest.raises(NotImplementedError):
        r.run("foo")
    with pytest.raises(NotImplementedError):
        # This is implemented but calls `.run()`
        r.run_bash("foo")
    with pytest.raises(NotImplementedError):
        r.user("nonroot")
    with pytest.raises(NotImplementedError):
        r.workdir("/opt/")


# TODO: add many tests for `indent`.



import pytest

from neurodocker.reproenv.exceptions import RendererError
from neurodocker.reproenv.renderers import SingularityRenderer
from neurodocker.reproenv.template import Template
from neurodocker.reproenv.tests.utils import prune_rendered


def test_singularity_renderer_add_template():
    s = SingularityRenderer("apt")

    d = {
        "name": "foobar",
        "binaries": {
            "urls": {"1.0.0": "foobar"},
            "env": {"foo": "bar"},
            "instructions": (
                "{{self.install_dependencies()}}\necho hello {{ self.myname }}"
            ),
            "arguments": {"required": ["myname"], "optional": {}},
            "dependencies": {
                "apt": ["curl", "wget"],
                "debs": [],
                "yum": ["python", "wget"],
            },
        },
    }

    with pytest.raises(
        RendererError, match="template must be an instance of 'Template' but got"
    ):
        s.add_template(d, method="binaries")

    # Invalid method
    with pytest.raises(
        RendererError, match="method must be 'binaries', 'source' but got 'fakemethod'"
    ):
        s.add_template(Template(d, binaries_kwds=dict(myname="f")), method="fakemethod")

    # Test apt.
    s = SingularityRenderer("apt")
    s.add_template(Template(d, binaries_kwds=dict(myname="Bjork")), method="binaries")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
%environment
export foo="bar"

%post
apt-get update -qq
apt-get install -y -q --no-install-recommends \\
    curl \\
    wget
rm -rf /var/lib/apt/lists/*
echo hello Bjork"""
    )

    d = {
        "name": "baz",
        "binaries": {
            "urls": {"1.0.0": "foobar"},
            "env": {"foo": "bar"},
            "instructions": (
                "{{self.install_dependencies()}}\necho hello {{ self.myname }}"
            ),
            "arguments": {"required": [], "optional": {"myname": "foo"}},
            "dependencies": {"apt": ["curl wget"], "debs": [], "yum": ["python wget"]},
        },
    }

    s = SingularityRenderer("apt")
    s.add_template(Template(d), method="binaries")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
%environment
export foo="bar"

%post
apt-get update -qq
apt-get install -y -q --no-install-recommends \\
    curl wget
rm -rf /var/lib/apt/lists/*
echo hello foo"""
    )


def test_singularity_render_from_instance_methods():
    s = SingularityRenderer("apt")
    s.from_("alpine")
    s.copy(["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
Bootstrap: docker
From: alpine

%files
foo/bar/baz.txt /opt/
foo/baz/cat.txt /opt/

%post"""
    )

    s = SingularityRenderer("apt")
    s.from_("alpine")
    s.copy(["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/")
    s.env(FOO="BAR")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
Bootstrap: docker
From: alpine

%files
foo/bar/baz.txt /opt/
foo/baz/cat.txt /opt/

%environment
export FOO="BAR"

%post"""
    )

    # Label
    s = SingularityRenderer("apt")
    s.from_("alpine")
    s.copy(["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/")
    s.env(FOO="BAR")
    s.label(ORG="BAZ")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
Bootstrap: docker
From: alpine

%files
foo/bar/baz.txt /opt/
foo/baz/cat.txt /opt/

%environment
export FOO="BAR"

%post


%labels
ORG BAZ"""
    )

    # Run
    s = SingularityRenderer("apt")
    s.from_("alpine")
    s.copy(["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/")
    s.env(FOO="BAR")
    s.label(ORG="BAZ")
    s.labels({'org.test.label': 'BAX'})
    s.run("echo foobar")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
Bootstrap: docker
From: alpine

%files
foo/bar/baz.txt /opt/
foo/baz/cat.txt /opt/

%environment
export FOO="BAR"

%post
echo foobar

%labels
ORG BAZ
org.test.label BAX"""
    )

    # User
    s = SingularityRenderer("apt")
    s.from_("alpine")
    s.copy(["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/")
    s.env(FOO="BAR")
    s.label(ORG="BAZ")
    s.run("echo foobar")
    s.user("nonroot")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
Bootstrap: docker
From: alpine

%files
foo/bar/baz.txt /opt/
foo/baz/cat.txt /opt/

%environment
export FOO="BAR"

%post
echo foobar

test "$(getent passwd nonroot)" \\
|| useradd --no-user-group --create-home --shell /bin/bash nonroot


su - nonroot

%labels
ORG BAZ"""
    )

    # nonroot user
    s = SingularityRenderer("apt")
    s.from_("alpine")
    s.copy(["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/")
    s.env(FOO="BAR")
    s.label(ORG="BAZ")
    s.run("echo foobar")
    s.user("nonroot")
    s.workdir("/opt/foo")
    s.user("root")
    s.user("nonroot")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
Bootstrap: docker
From: alpine

%files
foo/bar/baz.txt /opt/
foo/baz/cat.txt /opt/

%environment
export FOO="BAR"

%post
echo foobar

test "$(getent passwd nonroot)" \\
|| useradd --no-user-group --create-home --shell /bin/bash nonroot


su - nonroot

mkdir -p /opt/foo
cd /opt/foo

su - root

su - nonroot

%labels
ORG BAZ"""
    )

    # run bash
    s = SingularityRenderer("apt")
    s.from_("alpine")
    s.copy(["foo/bar/baz.txt", "foo/baz/cat.txt"], "/opt/")
    s.env(FOO="BAR")
    s.label(ORG="BAZ")
    s.run("echo foobar")
    s.user("nonroot")
    s.workdir("/opt/foo")
    s.user("root")
    s.user("nonroot")
    s.run_bash("source activate")
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
Bootstrap: docker
From: alpine

%files
foo/bar/baz.txt /opt/
foo/baz/cat.txt /opt/

%environment
export FOO="BAR"

%post
echo foobar

test "$(getent passwd nonroot)" \\
|| useradd --no-user-group --create-home --shell /bin/bash nonroot


su - nonroot

mkdir -p /opt/foo
cd /opt/foo

su - root

su - nonroot

bash -c 'source activate'

%labels
ORG BAZ"""
    )

    s = SingularityRenderer("apt")
    s.from_("debian:buster-slim")
    s.entrypoint(["echo", "foobar baz"])
    rendered = str(s)
    rendered = prune_rendered(rendered).strip()
    assert (
        rendered
        == """\
Bootstrap: docker
From: debian:buster-slim

%post


%runscript
echo foobar baz"""
    )



"""Remove all files under a directory but not caught by `reprozip trace`."""

from pathlib import Path
import typing as ty

import yaml

# TODO: if removing all of the files in a directory, print that directory with a *.
# pip does this in `pip uninstall`.


def _in_docker() -> bool:
    if not Path("/proc/1/cgroup").is_file():
        return False
    # TODO: this is not a long-term solution.
    return Path("/.dockerenv").is_file()


def main(
    *,
    yaml_file: ty.Union[str, Path],
    directories_to_prune: ty.Union[ty.List[str], ty.List[Path]],
):

    if not _in_docker():
        raise RuntimeError(
            "Not running in a Docker container. This script should only be used within"
            " a container."
        )

    yaml_file = Path(yaml_file)
    if yaml_file.name != "config.yml":
        raise ValueError("File should be named `config.yml`")
    with yaml_file.open(mode="r") as f:
        config = yaml.load(f, Loader=yaml.SafeLoader)

    # Paths caught by `reprozip trace`. These can be directories or files.
    # There could potentially be other information (e.g., input/output,
    # packages), but we are only interested in the contents of this key.
    files_to_keep = config["other_files"]
    files_to_keep = {Path(p) for p in files_to_keep}

    if isinstance(directories_to_prune, (Path, str)):
        directories_to_prune = [directories_to_prune]

    directories_to_prune = [Path(d) for d in directories_to_prune]

    # TODO: consider moving this check to the `neurodocker minify` cli. perhaps make a
    # new click paramtype that checks if the directory exists in the container.
    for d in directories_to_prune:
        if not d.is_dir():
            raise ValueError(f"Directory does not exist: {d}")

    all_files: ty.Set[Path] = set()
    for d in directories_to_prune:
        all_files.update(Path(d).rglob("*"))

    files_to_remove = all_files - files_to_keep
    files_to_remove = {f for f in files_to_remove if f.is_file()}

    with open(
        "/tmp/neurodocker-reprozip-trace/TO_DELETE.list", mode="w", encoding="utf-8"
    ) as f:
        n_errored = 0
        for ff in sorted(files_to_remove):
            try:
                print(ff, file=f)
            except UnicodeEncodeError:
                n_errored += 1

    print(f"++ skipping {n_errored} files due to unicode encoding errors")
    print(f"++ found {len(files_to_remove) - n_errored} files to remove")


if __name__ == "__main__":
    from argparse import ArgumentParser

    p = ArgumentParser(description=__doc__)
    p.add_argument(
        "--config-file", required=True, help="`config.yml` file from `reprozip trace`."
    )
    p.add_argument(
        "--dirs-to-prune",
        required=True,
        nargs="+",
        help="Directories to prune. Data will be lost in these directories.",
    )
    args = p.parse_args()

    main(yaml_file=args.config_file, directories_to_prune=args.dirs_to_prune)



# see https://click.palletsprojects.com/en/7.x/advanced/#forwarding-unknown-options

# TODO: consider using https://github.com/click-contrib/click-option-group to create
# groups of options in the cli. Could be helpful to separate the options.

# TODO: add a dedicated class for key=value in the eat-all class.

import json as json_lib
from pathlib import Path
import sys
import typing as ty

import click

from neurodocker.reproenv.renderers import _Renderer
from neurodocker.reproenv.renderers import DockerRenderer
from neurodocker.reproenv.renderers import SingularityRenderer
from neurodocker.reproenv.state import get_template
from neurodocker.reproenv.state import register_template
from neurodocker.reproenv.state import registered_templates
from neurodocker.reproenv.state import registered_templates_items
from neurodocker.reproenv.template import Template
from neurodocker.reproenv.types import allowed_pkg_managers


class GroupAddCommonParamsAndRegisteredTemplates(click.Group):
    """Subclass of `click.Group` that adds parameters common to `reproenv generate`
    commands, registers templates, and adds parameters to render templates.

    Note: Commands that are part of this group must be called using the group. This is
    because the `.get_command()` method of this group adds required parameters.
    """

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.params = [
            click.Option(
                ["--template-path"],
                multiple=True,
                envvar="REPROENV_TEMPLATE_PATH",
                show_envvar=True,
                help="Path to directories with templates to register",
                type=click.Path(exists=True, file_okay=False, dir_okay=True),
            )
        ]

    def get_command(self, ctx: click.Context, name: str) -> ty.Optional[click.Command]:
        command = self.commands.get(name)
        if command is None:
            return command  # return immediately to error can be logged

        # This is only set if a subcommand is called. Calling --help on the group
        # does not set --template-path.
        template_path: ty.Tuple[str] = ctx.params.get("template_path", tuple())
        yamls: ty.List[Path] = []
        for p in template_path:
            path = Path(p)
            for pattern in ("*.yaml", "*.yml"):
                yamls.extend(path.glob(pattern))
        # TODO: log warning if no yamls are found?
        for path in yamls:
            _ = register_template(path)

        params: ty.List[click.Parameter] = [
            click.Option(
                ["-p", "--pkg-manager"],
                type=click.Choice(list(allowed_pkg_managers), case_sensitive=False),
                required=True,
                multiple=False,
                help="System package manager",
            )
        ]
        params = _get_common_renderer_params()
        params += _get_params_for_registered_templates()
        command.params += params
        return command


# https://stackoverflow.com/a/65744803/5666087
class OrderedParamsCommand(click.Command):
    """Subclass of `click.Command` that maintains the order of user-provided
    parameters.
    """

    def parse_args(self, ctx: click.Context, args: ty.List[str]):
        self._options: ty.List[ty.Tuple[click.Parameter, ty.Any]] = []
        # run the parser for ourselves to preserve the passed order
        parser = self.make_parser(ctx)
        param_order: ty.List[click.Parameter]
        opts, _, param_order = parser.parse_args(args=list(args))
        for param in param_order:
            # We need the parameter name... so if it's None, let's panic.
            if param.name is None:
                raise ValueError(f"parameter name is None: {param}")
            value = opts[param.name]
            # If we have multiple values, take the first one. We do this before type
            # casting, because type casting for some reason brings all of the given
            # values for a parameter into the container. Not sure why, but perhaps it
            # has to do with the click.Context object.
            if isinstance(value, list):
                value = value.pop(0)
            if param.multiple:
                # If the value is supposed to be in a tuple, put it back in a tuple.
                value = (value,)
            value = param.type_cast_value(ctx, value)
            if isinstance(value, tuple):
                value = value[0]
            self._options.append((param, value))

        # return "normal" parse results
        return super().parse_args(ctx, args)


# https://stackoverflow.com/a/48394004/5666087
class OptionEatAll(click.Option):
    """Subclass of `click.Option` that allows for an arbitrary number of options.

    The behavior is similar to `nargs="*"` in argparse.
    """

    def __init__(self, *args, **kwargs):
        nargs = kwargs.pop("nargs", -1)
        assert nargs == -1, "nargs, if set, must be -1 not {}".format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):
        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            # grab everything up to the next option
            while state.rargs and not done:
                for prefix in self._eat_all_parser.prefixes:
                    if state.rargs[0].startswith(prefix):
                        done = True
                if not done:
                    value.append(state.rargs.pop(0))
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


class KeyValuePair(click.ParamType):
    """Type that accepts key=value pairs and converts to tuples of (key, value) if
    not multiple and ((key1, value1), (key2, value2)...) if multiple.
    """

    name = "key=value"

    def convert(self, value, param, ctx):
        def fn(v: str):
            # Limiting to one split allows values to include =.
            # For example, package versions in miniconda can include =,
            # like conda_install="python=3.9".
            strs = v.split("=", maxsplit=1)
            if len(strs) != 2:
                self.fail("expected string in format 'key=value'", param, ctx)
            k, v = strs
            return k, v

        # This might be a tuple or a list if using OptionEatAll.
        if isinstance(value, (list, tuple)):
            return tuple(map(fn, value))
        else:
            return fn(value)


def _get_common_renderer_params() -> ty.List[click.Parameter]:
    params: ty.List[click.Parameter] = [
        click.Option(
            ["-p", "--pkg-manager"],
            type=click.Choice(list(allowed_pkg_managers), case_sensitive=False),
            required=True,
            multiple=False,
            help="System package manager",
        ),
        click.Option(
            ["-b", "--base-image", "from_"],
            required=True,
            multiple=True,
            help="Base image",
        ),
        click.Option(
            ["--arg"],
            type=KeyValuePair(),
            multiple=True,
            help="Build-time variables (do not persist after container is built)",
        ),
        OptionEatAll(
            ["--copy"],
            multiple=True,
            help=(
                "Copy files into the container. Provide at least two paths."
                " The last path is always the destination path in the container."
            ),
        ),
        OptionEatAll(
            ["--env"],
            multiple=True,
            type=KeyValuePair(),
            help="Set persistent environment variables",
        ),
        OptionEatAll(
            ["--entrypoint"],
            multiple=True,
            help="Set entrypoint of the container",
        ),
        OptionEatAll(
            ["--install"],
            multiple=True,
            help="Install packages with system package manager",
        ),
        OptionEatAll(
            ["--label"],
            multiple=True,
            type=KeyValuePair(),
            help="Set labels on the container",
        ),
        click.Option(["--run"], multiple=True, help="Run commands in /bin/sh"),
        click.Option(
            ["--run-bash"], multiple=True, help="Run commands in a bash shell"
        ),
        click.Option(
            ["--user"],
            multiple=True,
            help="Switch to a different user (create user if it does not exist)",
        ),
        click.Option(["--workdir"], multiple=True, help="Set the working directory"),
        click.Option(["--yes"], is_flag=True, help="Reply yes to all prompts."),
        click.Option(
            ["--json"],
            is_flag=True,
            help=(
                "Output instructions as JSON. This can be used to generate Dockerfiles"
                " or Singularity recipes with Neurodocker."
            ),
        ),
    ]
    return params


def _create_help_for_template(template: Template) -> str:
    """Return a string help message for a template.

    This help message lists available installation methods and required and optional
    parameters for each method.
    """
    methods = []
    if template.binaries is not None:
        methods.append("binaries")
    if template.source is not None:
        methods.append("source")
    h = f"\b\nAdd {template.name}\n  method=[{'|'.join(methods)}]"
    for method in methods:
        h += f"\n  options for method={method}"
        for arg in getattr(template, method).required_arguments:
            h += f"\n    - {arg} [required]"
            # TODO: should we only include versions if using binaries?
            if arg == "version" and method == "binaries":
                h += f"""\n        version=[{'|'.join(
                    sorted(getattr(template, method).versions, reverse=True))}]"""
        for arg, default in getattr(template, method).optional_arguments.items():
            h += f"\n    - {arg} [default: {default}]"
    if template.alert:
        h += f"\n**Note**: {template.alert}"
    return h


def _get_params_for_registered_templates() -> ty.List[click.Parameter]:
    """Return list of click parameters for registered templates."""
    params: ty.List[click.Parameter] = []
    names_tmpls = list(registered_templates_items())
    names_tmpls.sort(key=lambda r: r[0])  # sort by name
    for name, tmpl in names_tmpls:
        hlp = _create_help_for_template(Template(tmpl))
        param = OptionEatAll(
            [f"--{name.lower()}"], type=KeyValuePair(), multiple=True, help=hlp
        )
        params.append(param)
    return params


def _params_to_renderer_dict(ctx: click.Context, pkg_manager) -> dict:
    """Return dictionary compatible with compatible with `_Renderer.from_dict()`."""
    renderer_dict = {"pkg_manager": pkg_manager, "instructions": []}
    cmd = ctx.command
    cmd = ty.cast(OrderedParamsCommand, cmd)
    for param, value in cmd._options:
        d = _get_instruction_for_param(ctx=ctx, param=param, value=value)
        # TODO: what happens if `d is None`?
        if d is not None:
            renderer_dict["instructions"].append(d)
    if not renderer_dict["instructions"]:
        ctx.fail("not enough instructions to generate a container specification")
    return renderer_dict


def _get_instruction_for_param(
    ctx: click.Context, param: click.Parameter, value: ty.Any
):
    # TODO: clean this up.
    d = None
    if param.name == "from_":
        d = {"name": param.name, "kwds": {"base_image": value}}
    # arg
    elif param.name == "arg":
        assert len(value) == 2, "expected key=value pair for --arg"
        k, v = value
        d = {"name": param.name, "kwds": {"key": k, "value": v}}
    # copy
    elif param.name == "copy":
        assert len(value) > 1, "expected at least two values for --copy"
        source, destination = list(value[:-1]), value[-1]
        d = {"name": param.name, "kwds": {"source": source, "destination": destination}}
    # env
    elif param.name == "env":
        value = dict(value)
        d = {"name": param.name, "kwds": {**value}}
    # entrypoint
    elif param.name == "entrypoint":
        if isinstance(value, str):
            value = [value]
        else:
            value = list(value)
        d = {"name": param.name, "kwds": {"args": value}}
    # install
    elif param.name == "install":
        opts = None
        if isinstance(value, tuple):
            value = list(value)
        # Special case. Set `opts` if provided.
        for idx, val in enumerate(value):
            if val.startswith("opts="):
                opts = value.pop(idx)
                opts = opts[slice(len("opts="), None)]  # remove leading opts=
        if not value:
            ctx.fail("no packages provided to install")
        d = {"name": param.name, "kwds": {"pkgs": value, "opts": opts}}
    # label
    elif param.name == "label":
        value = dict(value)
        d = {"name": param.name, "kwds": {**value}}
    # run
    elif param.name == "run":
        d = {"name": param.name, "kwds": {"command": value}}
    # run_bash
    elif param.name == "run_bash":
        d = {"name": param.name, "kwds": {"command": value}}
    # user
    elif param.name == "user":
        d = {"name": param.name, "kwds": {"user": value}}
    # workdir
    elif param.name == "workdir":
        d = {"name": param.name, "kwds": {"path": value}}
    # probably a registered template?
    else:
        # We need the parameter name... so if it's None, let's panic.
        if param.name is None:
            raise ValueError(f"parameter name is None: {param}")
        if param.name.lower() in registered_templates():
            tmpl_name = param.name.lower()
            value = dict(value)
            d = {"name": tmpl_name, "kwds": dict(value)}
            # If the template has an alert, prompt the user for confirmation.
            tmpl = Template(get_template(tmpl_name))
            if tmpl.alert:
                # If user provides --yes flag, print message but do not ask for
                # confirmation.
                yes = ctx.params.get("yes")
                if yes:
                    click.secho(tmpl.alert, fg="yellow", err=True)
                else:
                    # TODO: add color to this to make it visible, perhaps yellow. But
                    # there is no color option in `click.confirm`.
                    click.confirm(f"{tmpl.alert} Proceed?", abort=True, err=True)
        else:
            # TODO: should we do anything special with unknown options? Probably log it.
            pass
    return d


@click.group(cls=GroupAddCommonParamsAndRegisteredTemplates)
def generate(*, template_path):
    """Generate a container."""
    pass


def _base_generate(
    ctx: click.Context, renderer: ty.Type[_Renderer], pkg_manager: str, **kwds
):
    """Function that does all of the work of `generate docker` and
    `generate singularity`. The difference between those two is the renderer used.
    """
    renderer_dict = _params_to_renderer_dict(ctx=ctx, pkg_manager=pkg_manager)
    r = renderer.from_dict(renderer_dict)

    # Print the instructions in JSON if that's what the user wants.
    # We get the JSON instructions from the renderer itself -- rather than the
    # `renderer_dict` -- because the renderer instructions containers (mostly)
    # reproducible commands. A user might pass `--fsl version=6.0.4`, and the JSON will
    # include the commands associated with that version of FSL. If we had printed the
    # `renderer_dict`, then the dictionary would include `fsl` and version 6.0.4, but
    # the behavior of that installation could change if there is a change in the
    # template for FSL in neurodocker.
    json = kwds.get("json", False)
    if json:
        click.echo(r.to_json())
        ctx.exit(0)

    output = str(r)
    click.echo(output)


@generate.command(cls=OrderedParamsCommand)
@click.pass_context
def docker(ctx: click.Context, pkg_manager: str, **kwds):
    """Generate a Dockerfile."""
    _base_generate(
        ctx=ctx,
        renderer=DockerRenderer,
        pkg_manager=pkg_manager,
        **kwds,
    )


@generate.command(cls=OrderedParamsCommand)
@click.pass_context
def singularity(ctx: click.Context, pkg_manager: str, **kwds):
    """Generate a Singularity recipe."""
    _base_generate(
        ctx=ctx,
        renderer=SingularityRenderer,
        pkg_manager=pkg_manager,
        **kwds,
    )


@click.command()
@click.argument(
    "container_type",
    required=True,
    type=click.Choice(["docker", "singularity"], case_sensitive=False),
)
@click.argument(
    "input",
    type=click.File("r"),
    default=sys.stdin,
)
def genfromjson(*, container_type: str, input: ty.IO):
    """Generate a container from a ReproEnv JSON file.

    INPUT is standard input by default or a path to a JSON file.
    """
    d = json_lib.load(input)

    renderer: ty.Type[_Renderer]
    if container_type.lower() == "docker":
        renderer = DockerRenderer
    elif container_type.lower() == "singularity":
        renderer = SingularityRenderer

    r = renderer.from_dict(d)
    spec = str(r)
    click.echo(spec)



import pytest

from neurodocker.reproenv import exceptions
from neurodocker.reproenv import template
from neurodocker.reproenv import types


def test_template():
    d = {
        "name": "foobar",
        "binaries": {
            "urls": {"v1": "foo"},
            "env": {"baz": "cat", "boo": "123"},
            "instructions": "echo hi there\n{{ self.baz }}",
            "arguments": {"required": ["baz"], "optional": {}},
            "dependencies": {"apt": ["curl"], "debs": ["foo"], "yum": ["curl"]},
        },
        "source": {
            "env": {"foo": "bar"},
            "instructions": "echo foo\n{{ self.boo }}",
            "arguments": {"required": [], "optional": {"boo": "baz"}},
            "dependencies": {"apt": ["curl"], "debs": [], "yum": []},
        },
    }
    # do not provide required kwds
    with pytest.raises(exceptions.TemplateKeywordArgumentError):
        template.Template(d).binaries.validate_kwds()
    # this is fine because source only has optional args
    template.Template(d).source.validate_kwds()
    # only provide optional kwd to source
    with pytest.raises(exceptions.TemplateKeywordArgumentError):
        template.Template(d, source_kwds=dict(boo="cat")).binaries.validate_kwds()
    # this is fine
    template.Template(d, source_kwds=dict(boo="cat")).source.validate_kwds()
    # provide unknown kwd
    with pytest.raises(exceptions.TemplateKeywordArgumentError):
        template.Template(d, source_kwds=dict(boop="cat")).source.validate_kwds()
    # provide empty dict
    with pytest.raises(exceptions.TemplateKeywordArgumentError):
        template.Template(
            d, binaries_kwds={}, source_kwds=dict(boop="cat")
        ).binaries.validate_kwds()
    # this is also bad
    with pytest.raises(exceptions.TemplateKeywordArgumentError):
        template.Template(
            d, binaries_kwds={}, source_kwds=dict(boop="cat")
        ).source.validate_kwds()
    # provide all kwds
    t = template.Template(d, binaries_kwds=dict(baz=1234), source_kwds=dict(boo="cat"))
    t.binaries.validate_kwds()
    t.source.validate_kwds()
    # do not provide optional kwds
    t = template.Template(d, binaries_kwds=dict(baz=1234))
    t.binaries.validate_kwds()
    t.source.validate_kwds()
    assert t.source.boo == d["source"]["arguments"]["optional"]["boo"]

    assert t.name == d["name"]
    assert t.binaries._template == d["binaries"]
    assert t.source._template == d["source"]
    assert t._template is not d  # check for deepcopy

    # only contains binaries
    t = template.Template(
        {
            "name": "foobar",
            "binaries": {
                "urls": {"v1": "foo"},
                "env": {"baz": "cat", "boo": "123"},
                "instructions": "echo hi there\n{{ self.baz }}",
                "arguments": {"required": ["baz"], "optional": {}},
                "dependencies": {"apt": ["curl"], "debs": ["foo"], "yum": ["curl"]},
            },
        },
        binaries_kwds=dict(baz="1234"),
    )
    assert t.source is None
    t.binaries.validate_kwds()

    t = template.Template(
        {
            "name": "foobar",
            "source": {
                "env": {"foo": "bar"},
                "instructions": "echo foo\n{{ self.boo }}",
                "arguments": {"required": [], "optional": {"boo": "baz"}},
                "dependencies": {"apt": ["curl"], "debs": [], "yum": []},
            },
        },
        source_kwds=dict(boo="baz"),
    )
    assert t.binaries is None
    t.source.validate_kwds()

    # invalid template, but otherwise ok
    with pytest.raises(exceptions.TemplateError):
        template.Template(
            {
                "source": {
                    "env": {"foo": "bar"},
                    "instructions": "echo foo\n{{ self.boo }}",
                    "arguments": {"required": [], "optional": {"boo": "baz"}},
                    "dependencies": {"apt": ["curl"], "dpkg": [], "yum": []},
                }
            },
            source_kwds=dict(boo="baz"),
        )


def test_installation_template_base():
    d: types.BinariesTemplateType = {
        "urls": {"1.0.0": "foobar"},
        "instructions": "foobar",
    }

    it = template._BaseInstallationTemplate(d)
    assert it._template == d
    assert it.env == {}
    assert it.instructions == d["instructions"]
    assert it.arguments == {}
    assert it.required_arguments == set()
    assert it.optional_arguments == {}
    assert it.dependencies("apt") == []
    assert it.dependencies("foobar") == []

    d: types.BinariesTemplateType = {
        "urls": {"1.0.0": "foobar"},
        "instructions": "hello {{ self.name }}",
        "arguments": {"required": ["name"], "optional": {"age": "50"}},
    }

    #
    # Test keyword arguments.
    #

    # missing required key
    with pytest.raises(exceptions.TemplateError, match="Missing required arguments"):
        template._BaseInstallationTemplate(d).validate_kwds()

    # bad key
    with pytest.raises(
        exceptions.TemplateError,
        match="Keyword argument provided is not specified in template",
    ):
        template._BaseInstallationTemplate(d, name="foo", unknown="val").validate_kwds()

    # optional key provided only
    with pytest.raises(exceptions.TemplateError, match="Missing required arguments"):
        template._BaseInstallationTemplate(d, age="42").validate_kwds()

    # versions
    d["arguments"]["required"] += ["version"]
    it = template._BinariesTemplate(d, name="didi", version="1.0.0", age=42)
    assert it.name == "didi"
    assert it.version == "1.0.0"
    # Values are all cast to string.
    assert it.age == "42"
    # invalid version - not found in urls
    with pytest.raises(exceptions.TemplateError):
        template._BinariesTemplate(d, version="2.0.0").validate_kwds()

    #
    # Source template
    #
    d: types.SourceTemplateType = {
        "instructions": "hello {{ self.name }}",
        "arguments": {"required": ["name"], "optional": {"age": "52"}},
    }
    it = template._SourceTemplate(d, name="foobar")
    assert it.versions == {"ANY"}

    d: types.SourceTemplateType = {
        "env": {"foo": "bar", "cat": "dog"},
        "instructions": "echo {{ name }}",
        "arguments": {
            "required": ["name"],
            "optional": {"age": "24", "height": "100 m"},
        },
        "dependencies": {"apt": ["curl"], "dpkg": [], "yum": ["python"]},
    }
    it = template._SourceTemplate(d, name="foobar", age=42, height=100)
    assert it._template == d
    assert it.env == {"foo": "bar", "cat": "dog"}
    assert it.instructions == d["instructions"]
    assert it.arguments == d["arguments"]
    assert it.required_arguments == set(d["arguments"]["required"])
    assert it.optional_arguments == d["arguments"]["optional"]
    assert it.dependencies("apt") == d["dependencies"]["apt"]
    # TODO: add dpkg
    assert it.dependencies("yum") == d["dependencies"]["yum"]
    assert it.dependencies("foobar") == []
    assert it._kwds == {"name": "foobar", "age": "42", "height": "100"}
    assert it.name == "foobar"
    assert it.age == "42"
    assert it.height == "100"


def test_template_alert():
    d: types.TemplateType = {
        "alert": "This is an alert!",
        "name": "testing",
        "binaries": {"urls": {"foo": "foo.baz.tar.gz"}, "instructions": "do nothing"},
    }
    tmpl = template.Template(d)
    assert tmpl.alert == d["alert"]

    del d["alert"]
    tmpl = template.Template(d)
    assert tmpl.alert == ""



"""Setup script for neurodocker."""

from setuptools import setup

import versioneer

version = versioneer.get_version()
cmdclass = versioneer.get_cmdclass()

setup(version=version, cmdclass=cmdclass)



from pathlib import Path

import pytest
import yaml

from neurodocker.reproenv import exceptions
from neurodocker.reproenv.state import _TemplateRegistry, _validate_template
from neurodocker.reproenv import types


def test_validate_template_invalid_templates():
    with pytest.raises(exceptions.TemplateError, match="'name' is a required property"):
        _validate_template({})

    with pytest.raises(
        exceptions.TemplateError, match="'binaries' is a required property"
    ):
        _validate_template({"name": "bar"})

    # missing 'name' top-level key
    with pytest.raises(exceptions.TemplateError, match="'name' is a required property"):
        _validate_template(
            {
                # "name": "foobar",
                "binaries": {"urls": {"1.0.0": "foobar.com"}, "instructions": "foobar"}
            }
        )

    # 'name' top-level value is not a string
    with pytest.raises(exceptions.TemplateError, match="1234 is not of type 'string'"):
        _validate_template(
            {
                "name": 1234,
                "binaries": {"urls": {"1.0.0": "foobar.com"}, "instructions": "foobar"},
            }
        )

    #
    # test of 'binaries' templates
    #

    # no instructions
    with pytest.raises(
        exceptions.TemplateError, match="'instructions' is a required property"
    ):
        _validate_template(
            {
                "name": "foobar",
                "binaries": {
                    "urls": {"1.0.0": "foobar.com"},
                    # "instructions": "foobar",
                },
            }
        )

    # malformed env
    with pytest.raises(
        exceptions.TemplateError,
        match="Invalid template: \\['foo'\\] is not of type 'string'.",
    ):
        _validate_template(
            {
                "name": "foobar",
                "binaries": {
                    "urls": {"1.0.0": "foobar.com"},
                    "env": {"foo": ["foo"]},
                    "instructions": "foobar",
                },
            }
        )

    # binaries but no urls
    with pytest.raises(exceptions.TemplateError, match="'urls' is a required property"):
        _validate_template(
            {
                "name": "foobar",
                "binaries": {
                    # "urls": {"1.0.0": "foobar.com"},
                    "instructions": "foobar"
                },
            }
        )

    # extra keys
    with pytest.raises(
        exceptions.TemplateError,
        match=(
            "Invalid template: Additional properties are not allowed \\('extra' was"
            " unexpected\\)"
        ),
    ):
        _validate_template(
            {
                "name": "foobar",
                "binaries": {
                    "urls": {"1.0.0": "foobar.com"},
                    "instructions": "foobar",
                    "extra": "",
                },
            }
        )

    # extra keys in dependencies
    with pytest.raises(
        exceptions.TemplateError,
        match=(
            "Invalid template: Additional properties are not allowed \\('fakemngr'"
            " was unexpected\\)."
        ),
    ):
        _validate_template(
            {
                "name": "foobar",
                "binaries": {
                    "urls": {"1.0.0": "foobar.com"},
                    "instructions": "foobar",
                    "dependencies": {"apt": [], "fakemngr": []},
                },
            }
        )

    # defines variable but does not indicate if optional or required
    # TODO

    #
    # test of 'source' templates
    #

    # urls should not be in source
    with pytest.raises(
        exceptions.TemplateError,
        match=(
            "Invalid template: Additional properties are not allowed \\('urls' was"
            " unexpected\\)."
        ),
    ):
        _validate_template(
            {
                "name": "foobar",
                "source": {"urls": {"1.0.0": "foobar.com"}, "instructions": "foobar"},
            }
        )

    # no instructions
    with pytest.raises(
        exceptions.TemplateError, match="'instructions' is a required property"
    ):
        _validate_template(
            {
                "name": "foobar",
                "source": {
                    # "instructions": "foobar",
                },
            }
        )

    # malformed env
    with pytest.raises(
        exceptions.TemplateError,
        match="Invalid template: \\['foo'\\] is not of type 'string'.",
    ):
        _validate_template(
            {
                "name": "foobar",
                "source": {"env": {"foo": ["foo"]}, "instructions": "foobar"},
            }
        )

    # extra keys
    with pytest.raises(
        exceptions.TemplateError,
        match=(
            "Invalid template: Additional properties are not allowed \\('extra' was"
            " unexpected\\)"
        ),
    ):
        _validate_template(
            {"name": "foobar", "source": {"instructions": "foobar", "extra": ""}}
        )

    # extra keys in dependencies
    with pytest.raises(
        exceptions.TemplateError,
        match=(
            "Invalid template: Additional properties are not allowed \\('fakemngr'"
            " was unexpected\\)."
        ),
    ):
        _validate_template(
            {
                "name": "foobar",
                "source": {
                    "instructions": "foobar",
                    "dependencies": {"apt": [], "fakemngr": []},
                },
            }
        )

    # defines variable but does not indicate if optional or required
    # TODO


def test_validate_template_valid_templates():
    # minimal templates
    _validate_template(
        {
            "name": "foobar",
            "binaries": {"urls": {"v1": "foo"}, "instructions": "foobar"},
        }
    )
    _validate_template({"name": "foobar", "source": {"instructions": "foobar"}})

    # bigger templates
    _validate_template(
        {
            "name": "foobar",
            "binaries": {
                "urls": {"v1": "foo"},
                "env": {"baz": "cat", "boo": "123"},
                "instructions": "echo hi there\n{{ self.install_dependencies() }}",
                "arguments": {"required": [], "optional": {}},
                "dependencies": {"apt": ["curl"], "debs": ["foo"], "yum": ["curl"]},
            },
            "source": {
                "env": {"foo": "bar"},
                "instructions": "echo foo\n{{ self.install_dependencies() }}",
                "arguments": {"required": [], "optional": {}},
                "dependencies": {"apt": ["curl"], "debs": [], "yum": []},
            },
        }
    )


def test_register(tmp_path: Path):
    _TemplateRegistry._reset()

    _one_test_template: types.TemplateType = {
        "name": "foobar",
        "binaries": {
            "urls": {"1.0.0": "foobar.com"},
            "env": {"foo": "bar"},
            "instructions": "foobar",
            "arguments": {"required": [], "optional": {}},
            "dependencies": {"apt": [], "debs": [], "yum": []},
        },
        "source": {
            "env": {"foo": "bar"},
            "instructions": "foobar",
            "arguments": {"required": [], "optional": {}},
            "dependencies": {"apt": [], "debs": [], "yum": []},
        },
    }

    # incorrect template type
    with pytest.raises(ValueError):
        _TemplateRegistry.register(path_or_template="", name="foobar")
    # empty template
    with pytest.raises(exceptions.TemplateError):
        _TemplateRegistry.register(path_or_template={}, name="foobar")

    # omit name
    with pytest.raises(ValueError):
        _TemplateRegistry.register(_one_test_template, name=None)
    with pytest.raises(ValueError):
        _TemplateRegistry.register(_one_test_template, name=None)

    # register using custom name
    name: str = "customname"
    _TemplateRegistry._reset()
    _TemplateRegistry.register(_one_test_template, name=name)
    assert _TemplateRegistry._templates[name] == _one_test_template
    assert not _TemplateRegistry._templates[name] is _one_test_template
    assert _TemplateRegistry.get(name) == _one_test_template
    assert _TemplateRegistry.get(name.upper()) == _one_test_template
    with pytest.raises(exceptions.TemplateNotFound):
        _TemplateRegistry.get(_one_test_template["name"])

    with pytest.raises(exceptions.TemplateNotFound):
        _TemplateRegistry.get("baz")

    yaml_path = tmp_path / "foo.yaml"
    with yaml_path.open("w") as f:
        yaml.dump(_one_test_template, f)

    # register using name defined in template
    name = _one_test_template["name"]
    _TemplateRegistry._reset()
    _TemplateRegistry.register(yaml_path)
    assert _TemplateRegistry._templates[name] == _one_test_template
    assert not _TemplateRegistry._templates[name] is _one_test_template
    assert _TemplateRegistry.get(name) == _one_test_template
    assert _TemplateRegistry.get(name.upper()) == _one_test_template

    # register using custom name
    name = "customfoobar"
    _TemplateRegistry._reset()
    _TemplateRegistry.register(yaml_path, name=name)
    assert _TemplateRegistry._templates[name] == _one_test_template
    assert not _TemplateRegistry._templates[name] is _one_test_template
    assert _TemplateRegistry.get(name) == _one_test_template
    assert _TemplateRegistry.get(name.upper()) == _one_test_template
    with pytest.raises(exceptions.TemplateNotFound):
        _TemplateRegistry.get(_one_test_template["name"])


def test_get():
    _TemplateRegistry._reset()

    with pytest.raises(exceptions.TemplateNotFound):
        _TemplateRegistry.get("foobar")

    d = {"name": "foo"}
    _TemplateRegistry._templates["foobar"] = d
    assert _TemplateRegistry.get("foobar") == d

    with pytest.raises(exceptions.TemplateNotFound):
        _TemplateRegistry.get("baz")


def test_keys():
    _TemplateRegistry._reset()
    assert _TemplateRegistry.keys() == set()

    name = "foo"
    _TemplateRegistry._templates[name] = {}
    assert _TemplateRegistry.keys() == {"foo"}



import click

from neurodocker import __version__
from neurodocker.cli.generate import generate
from neurodocker.cli.generate import genfromjson


@click.group()
@click.version_option(__version__, message="%(prog)s version %(version)s")
def cli():
    """Generate custom containers, and minify existing containers.

    The minify command is available only if Docker is installed and running.
    """


cli.add_command(generate)
cli.add_command(genfromjson)

# `docker-py` is required for minification but is not installed by default.
# We also pass if there is an error retrieving the Docker client.
# Say, for example, the Docker engine is not running (or it is not installed).
try:
    from neurodocker.cli.minify.trace import minify

    cli.add_command(minify)
except (ImportError, RuntimeError):
    # TODO: should we log a debug message? We don't even have a logger at the
    # time of writing, so probably not.
    pass



"""Template objects."""

from __future__ import annotations

import copy
import typing as ty

from neurodocker.reproenv.exceptions import TemplateKeywordArgumentError
from neurodocker.reproenv.state import _validate_template
from neurodocker.reproenv.types import _BinariesTemplateType
from neurodocker.reproenv.types import _SourceTemplateType
from neurodocker.reproenv.types import TemplateType


class Template:
    """Template object.

    This class makes it more convenient to work with templates. It also allows one to
    set keyword arguments for instances of templates. For example, if a template calls
    for an argument `version`, this class can be used to hold both the template and the
    value for `version`.

    Keywords are not validated during initialization. To validate keywords, use the
    instance method `self.binaries.validate_kwds()` and`self.source.validate_kwds()`.

    Parameters
    ----------
    template : TemplateType
        Dictionary that defines how to install software from pre-compiled binaries
        and/or from source.
    binaries_kwds : dict
        Keyword arguments to pass to the binaries section of the template. All keys and
        values must be strings.
    source_kwds : dict
        Keyword arguments passed to the source section of the template. All keys and
        values must be strings.
    """

    def __init__(
        self,
        template: TemplateType,
        binaries_kwds: ty.Mapping[str, str] = None,
        source_kwds: ty.Mapping[str, str] = None,
    ):
        # Validate against JSON schema. Registered templates were already validated at
        # registration time, but if we do not validate here, then in-memory templates
        # (ie python dictionaries) will never be validated.
        _validate_template(template)

        self._template = copy.deepcopy(template)
        self._binaries: ty.Optional[_BinariesTemplate] = None
        self._binaries_kwds = {} if binaries_kwds is None else binaries_kwds
        self._source: ty.Optional[_SourceTemplate] = None
        self._source_kwds = {} if source_kwds is None else source_kwds

        if "binaries" in self._template:
            self._binaries = _BinariesTemplate(
                self._template["binaries"], **self._binaries_kwds
            )
        if "source" in self._template:
            self._source = _SourceTemplate(
                self._template["source"], **self._source_kwds
            )

    @property
    def name(self) -> str:
        return self._template["name"]

    @property
    def binaries(self) -> ty.Union[None, _BinariesTemplate]:
        return self._binaries

    @property
    def source(self) -> ty.Union[None, _SourceTemplate]:
        return self._source

    @property
    def alert(self) -> str:
        """Return the template's `alert` property. Return an empty string if it does
        not exist.
        """
        return self._template.get("alert", "")


class _BaseInstallationTemplate:
    """Base class for installation template classes.

    This class and its subclasses make it more convenient to work with templates.
    It also allows one to set keyword arguments for instances of templates. For example,
    if a template calls for an argument `version`, this class can be used to hold both
    the template and the value for `version`.

    Parameters
    ----------
    template : BinariesTemplateType or SourceTemplateType
        Dictionary that defines how to install software from pre-compiled binaries or
        from source.
    kwds
        Keyword arguments to pass to the template. All values must be strings. Values
        that are not strings are cast to string.
    """

    def __init__(
        self,
        template: ty.Union[_BinariesTemplateType, _SourceTemplateType],
        **kwds: str,
    ) -> None:
        self._template = copy.deepcopy(template)
        # User-defined arguments that are passed to template at render time.
        for key, value in kwds.items():
            if not isinstance(value, str):
                kwds[key] = str(value)
        self._kwds = kwds

        # This is meant to be overwritten by renderers, so that self.pkg_manager can
        # be used in templates.
        self.pkg_manager = None

        # We cannot validate kwds immediately... The Renderer should not validate
        # immediately. It should validate only the installation method being used.
        self._set_kwds_as_attrs()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._template}, **{self._kwds})"

    def validate_kwds(self):
        """Raise `TemplateKeywordArgumentError` if keyword arguments to template are
        invalid.
        """
        # Check that all required keywords were provided by user.
        req_keys_not_found = self.required_arguments.difference(self._kwds)
        if req_keys_not_found:
            raise TemplateKeywordArgumentError(
                "Missing required arguments: '{}'.".format(
                    "', '".join(req_keys_not_found)
                )
            )

        # Check that unknown kwargs were not provided.
        all_kwds = self.required_arguments.union(self.optional_arguments.keys())
        unknown_kwds = set(self._kwds).difference(all_kwds)
        if unknown_kwds:
            raise TemplateKeywordArgumentError(
                "Keyword argument provided is not specified in template: '{}'.".format(
                    "', '".join(unknown_kwds)
                )
            )
        # Check that version is valid.
        if "version" in self.required_arguments:
            # At this point, we are certain "version" has been provided.
            v = self._kwds["version"]
            # Templates for builds from source have versions `{"ANY"}` because they can
            # ideally build any version.
            if (
                v not in self.versions
                # This indicates a source method.
                and self.versions != {"ANY"}
                # The presence of * in the list of binary urls indicates that any
                # version is allowed. It also suggests that the version passed by
                # the user is substituted into the URL for the binaries.
                # TODO: consider changing {"ANY"} to "*" in source methods.
                and "*" not in self.versions
            ):
                raise TemplateKeywordArgumentError(
                    "Unknown version '{}'. Allowed versions are '{}'.".format(
                        v, "', '".join(self.versions)
                    )
                )

    def _set_kwds_as_attrs(self):
        # Check that keywords do not shadow attributes of this object.
        shadowed = set(self._kwds).intersection(dir(self))
        if shadowed:
            raise TemplateKeywordArgumentError(
                "Invalid keyword arguments: '{}'. If these keywords are used by the"
                " template, then the template must be modified to use different"
                " keywords.".format("', '".join(shadowed))
            )
        # Set optional arguments to their default value, if the argument was not
        # provided.
        for k, v in self.optional_arguments.items():
            self._kwds.setdefault(k, v)
        # Set keywords as attributes to this object. This makes it easy to render
        # variables in the jinja template.
        for k, v in self._kwds.items():
            setattr(self, k, v)

    @property
    def template(self):
        return self._template

    @property
    def env(self) -> ty.Mapping[str, str]:
        return self._template.get("env", {})

    @property
    def instructions(self) -> str:
        return self._template.get("instructions", "")

    @property
    def arguments(self) -> ty.Mapping:
        return self._template.get("arguments", {})

    @property
    def required_arguments(self) -> ty.Set[str]:
        args = self.arguments.get("required", None)
        return set(args) if args is not None else set()

    @property
    def optional_arguments(self) -> ty.Dict[str, str]:
        args = self.arguments.get("optional", None)
        return args if args is not None else {}

    @property
    def versions(self) -> ty.Set[str]:
        raise NotImplementedError()

    def dependencies(self, pkg_manager: str) -> ty.List[str]:
        deps_dict = self._template.get("dependencies", {})
        # TODO: not sure why the following line raises a type error in mypy.
        return deps_dict.get(pkg_manager, [])  # type: ignore

    def install(self, pkgs: ty.List[str], opts: str = None) -> str:
        raise NotImplementedError(
            "This method is meant to be patched by renderer objects, so it can be used"
            " in templates and have access to the pkg_manager being used."
        )

    def install_dependencies(self, opts: str = None) -> str:
        raise NotImplementedError(
            "This method is meant to be patched by renderer objects, so it can be used"
            " in templates and have access to the pkg_manager being used."
        )


class _BinariesTemplate(_BaseInstallationTemplate):
    def __init__(self, template: _BinariesTemplateType, **kwds: str):
        super().__init__(template=template, **kwds)

    @property
    def urls(self) -> ty.Mapping[str, str]:
        # TODO: how can the code be changed so this cast is not necessary?
        self._template = ty.cast(_BinariesTemplateType, self._template)
        return self._template.get("urls", {})

    @property
    def versions(self) -> ty.Set[str]:
        # TODO: how can the code be changed so this cast is not necessary?
        self._template = ty.cast(_BinariesTemplateType, self._template)
        return set(self.urls.keys())


class _SourceTemplate(_BaseInstallationTemplate):
    def __init__(self, template: _SourceTemplateType, **kwds: str):
        super().__init__(template=template, **kwds)

    @property
    def versions(self) -> ty.Set[str]:
        return {"ANY"}
